"""Spatial kernel abstractions."""

import scipy
import numpy as np
import torch
from typing import Literal
from abc import ABC, abstractmethod
from smoother.weights import coordinate_to_weights_knn_sparse, sparse_weights_to_inv_cov

__all__ = ["SpatialCovKernel"]


class Kernel(ABC):
    """Abstract interface for kernel matrix representations.

    Implementations may store either dense kernels or low-rank factors, but
    must expose common operations required by SPLISOSM hypothesis tests.
    """

    @abstractmethod
    def __init__(self) -> None:
        """Initialize kernel-specific state."""

    @abstractmethod
    def realization(self) -> torch.Tensor:
        """Return the realized dense kernel matrix.

        Returns
        -------
        torch.Tensor
            Dense kernel matrix of shape ``(n_spots, n_spots)``.
        """

    @abstractmethod
    def xtKx(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the quadratic form ``x^T K x``.

        Parameters
        ----------
        x : torch.Tensor
            Input matrix of shape ``(n_spots, d)``.

        Returns
        -------
        torch.Tensor
            Quadratic form of shape ``(d, d)``.
        """

    @abstractmethod
    def eigenvalues(self, k: int | None = None) -> torch.Tensor:
        """Return leading eigenvalues of the kernel matrix.

        Parameters
        ----------
        k : int, optional
            Number of leading eigenvalues to return. If ``None``, returns all
            available eigenvalues.

        Returns
        -------
        torch.Tensor
            Eigenvalues sorted in descending order.
        """


class SpatialCovKernel(Kernel):
    """Graph-based spatial covariance kernel built from spot coordinates."""

    inv_cov: torch.Tensor
    """Sparse precision matrix (aka inverse spatial kernel) of shape (n_spots, n_spots)."""

    K_sp: torch.Tensor | None
    """
    Dense kernel matrix of shape (n_spots, n_spots) when full-rank storage is used.
    ``None`` when low-rank storage is used.
    """

    Q: torch.Tensor | None
    """
    Low-rank factor of shape (n_spots, rank) such that ``K_sp ~= Q @ Q.T``.
    ``None`` when full-rank storage is used.
    """

    def __init__(
        self,
        coords: np.ndarray | torch.Tensor,
        k_neighbors: int = 4,
        model: Literal["icar", "car", "isar", "sar"] = "icar",
        rho: float = 0.99,
        standardize_cov: bool = True,
        centering: bool = False,
        approx_rank: int | None = None,
    ) -> None:
        """
        Parameters
        ----------
        coords
            Spot coordinates of shape (n_spots, 2).
        k_neighbors
            Number of nearest neighbors used to build the graph.
        model
            Spatial process model for inverse covariance construction. Supported
            values are ``'icar'``, ``'car'``, ``'isar'``, and ``'sar'``.
        rho
            Spatial autocorrelation coefficient.
        standardize_cov
            If True, scales covariance to unit marginal variance.
        centering
            If True, applies centering so row/column sums are approximately
            zero.
        approx_rank
            If provided, computes and stores a rank-`approx_rank` factorization.
        """
        # store the configurations
        self._configs = {
            "k_neighbors": k_neighbors,
            "model": model,
            "rho": rho,
            "standardize_cov": standardize_cov,
            "centering": centering,
            "approx_rank": approx_rank,
        }

        # calculate the sparse inverse spatial covariance matrix from KNN graph
        swm = coordinate_to_weights_knn_sparse(
            coords, k=k_neighbors, symmetric=True, row_scale=False
        )
        inv_cov = sparse_weights_to_inv_cov(
            swm, model=model, rho=rho, standardize=False, return_sparse=True
        ).coalesce()  # (n_spots, n_spots), sparse
        self.inv_cov = inv_cov

        if approx_rank is None:
            # compute the (N, N) dense spatial covariance matrix
            # torch.cholesky() is faster than scipy.sparse.linalg.inv()
            cov_sp = torch.cholesky_inverse(torch.linalg.cholesky(inv_cov.to_dense()))

            if (
                standardize_cov
            ):  # standardize the variance of the covariance matrix to one
                inv_sds = torch.diagflat(torch.diagonal(cov_sp) ** (-0.5))
                cov_sp = inv_sds @ cov_sp @ inv_sds

            if centering:  # center the kernel matrix
                cov_sp = cov_sp - cov_sp.mean(dim=0, keepdim=True)
                cov_sp = cov_sp - cov_sp.mean(dim=1, keepdim=True)

            # store the dense spatial kernel matrix
            self.K_sp = cov_sp  # (n_spots, n_spots)
            self._rank = cov_sp.shape[0]  # full rank

            # compute eigenvalues and eigenvectors in later stages
            self.K_eigvals = None  # eigenvalues of the kernel matrix
            self.K_eigvecs = None  # eigenvectors of the kernel matrix
            self.Q = None  # low-rank approximation of the kernel matrix K_sp = Q @ Q^T
        else:
            # compute and store the low-rank approximation of the kernel matrix
            if approx_rank > inv_cov.shape[0]:
                raise ValueError("`approx_rank` must not exceed the number of spots.")

            # normalize the inverse covariance matrix by degree such that the diagonal entries are 1
            # inv_cov = D^(-1/2) @ inv_cov @ D^(-1/2)
            # this will update self.inv_cov
            degrees = swm.sum(dim=1).to_dense()  # (n_spots,)
            inv_cov.values()[:] = inv_cov.values() / degrees[inv_cov.indices()[0]].pow(
                0.5
            )  # (n_spots, n_spots)
            inv_cov.values()[:] = inv_cov.values() / degrees[inv_cov.indices()[1]].pow(
                0.5
            )  # (n_spots, n_spots)

            # first convert torch sparse tensor to scipy coo matrix
            indices = inv_cov.indices().numpy()
            values = inv_cov.values().numpy()
            shape = inv_cov.shape
            coo_mtx = scipy.sparse.coo_matrix(
                (values, (indices[0], indices[1])), shape=shape
            )

            # then compute the smallest k eigenvalues of the inverse covariance matrix
            eigvals, eigvecs = scipy.sparse.linalg.eigsh(
                coo_mtx, k=approx_rank, which="SM"
            )

            # which is equivalent to the largest k eigenvalues of the covariance matrix
            eigvals = torch.from_numpy(eigvals + 1e-6).pow(
                -1
            )  # sorted in descending order
            eigvecs = torch.from_numpy(eigvecs)

            # store the eigenvalues and eigenvectors
            self.cov_eigvals = eigvals
            self.cov_eigvecs = eigvecs

            # the low-rank approximation of the kernel matrix is K_sp = Q @ Q^T
            Q = eigvecs @ torch.diag(eigvals.pow(0.5))  # (n_spots, rank)

            # scale K_sp to have unit variance (diagonal entries)
            if standardize_cov:
                # after scaling, cov_sp and Q @ Q.T would have different eigenvalues
                # due to the low-rank approximation
                _sd = (Q.pow(2).sum(dim=1) + 1e-6).pow(0.5)  # (n_spots,)
                Q = Q / _sd[:, None]

            # center K_sp to have zero row and column sums
            if centering:
                Q = Q - Q.mean(dim=0, keepdim=True)

            # compute kernel eigenvalues from Q
            K_eigvals = torch.linalg.eigvalsh(Q.T @ Q)  # (rank,), ascending order

            # sort the eigenvalues in descending order
            idx = K_eigvals.argsort(descending=True)
            self.K_eigvals = K_eigvals[idx]

            # store the low-rank approximation of the kernel matrix
            self.Q = Q
            self.K_sp = None  # K_sp = Q @ Q^T
            self._rank = self.Q.shape[1]

    def shape(self) -> tuple[int, int]:
        """Return kernel shape.

        Returns
        -------
        tuple of int
            Pair (n_spots, n_spots).
        """
        return self.inv_cov.shape

    def rank(self) -> int:
        """Return effective rank of the kernel representation.

        Returns
        -------
        int
            Kernel rank.
        """
        return self._rank

    def realization(self) -> torch.Tensor:
        """Return the realized dense kernel matrix.

        Returns
        -------
        torch.Tensor
            Dense matrix of shape (n_spots, n_spots).
        """
        if self.K_sp is not None:
            return self.K_sp
        else:
            return self.Q @ self.Q.t()

    def xtKx(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the kernel quadratic form for multivariate inputs.

        Parameters
        ----------
        x : torch.Tensor
            Input matrix of shape (n_spots, d).

        Returns
        -------
        torch.Tensor
            Quadratic form of shape (d, d).
        """
        if self.K_sp is not None:  # use the full rank kernel matrix
            return x.t() @ self.K_sp @ x
        else:  # use the low-rank approximation
            xtQ = x.t() @ self.Q  # (d, rank)
            return xtQ @ xtQ.t()  # (d, d)

    def eigenvalues(self, k: int | None = None) -> torch.Tensor:
        """Return leading eigenvalues of the kernel matrix.

        Parameters
        ----------
        k
            Number of leading eigenvalues to return. If None, all
            available eigenvalues are returned.

        Returns
        -------
        torch.Tensor
            Eigenvalues sorted in descending order.
        """
        if self.K_eigvals is None:
            # compute and store the eigendecomposition of self.K_sp
            self.eigendecomposition()

        # return the top k largest eigenvalues
        if k is None:
            return self.K_eigvals
        else:
            return self.K_eigvals[:k]

    def eigendecomposition(self) -> None:
        """Compute and store eigendecomposition of the dense kernel.

        Raises
        ------
        ValueError
            If only low-rank kernel factors are stored.
        """
        if self.K_sp is None:
            raise ValueError(
                "Dense kernel storage is required for eigendecomposition()."
            )

        eigvals, eigvecs = np.linalg.eigh(self.K_sp.numpy())

        # sort the eigenvalues and eigenvectors in descending order
        idx = eigvals.argsort()[::-1]
        self.K_eigvals = torch.from_numpy(eigvals[idx])
        self.K_eigvecs = torch.from_numpy(eigvecs[:, idx])
        # Q = eigvecs @ np.diag(eigvals ** 0.5) # (n_spots, n_spots)
        # self.Q = torch.from_numpy(Q)[:, idx]



import unittest
import numpy as np
import torch
import itertools
from splisosm.kernel import SpatialCovKernel
from splisosm.utils import get_cov_sp


class TestSpatialCovKernel(unittest.TestCase):

    def setUp(self):
        # Simulate 80x80 grid coordinates
        x = np.linspace(0, 1, 30)
        y = np.linspace(0, 1, 30)
        self.n_spots = 900
        self.coords = np.array(list(itertools.product(x, y)))
        self.H = torch.eye(900) - 1 / 900

    def test_full_rank_kernel(self):
        # spatial kernel from the icar prior directly
        cov_sp1 = get_cov_sp(self.coords)
        K1 = self.H @ cov_sp1 @ self.H
        L1 = torch.linalg.eigvalsh(K1)  # sorted in ascending order
        self.assertTrue(L1.numpy()[1] > 0)  # check positive semi-definite

        # spatial kernel from the icar prior using the kernel class
        cov_sp2 = SpatialCovKernel(self.coords, approx_rank=None, centering=True)
        self.assertTrue(cov_sp2.rank() == 900)
        self.assertTrue(cov_sp2.shape() == (900, 900))

        # check eigenvalues
        K2 = cov_sp2.realization()
        L2 = cov_sp2.eigenvalues()
        L2 = torch.flip(L2, dims=(0,))  # sort in descending order
        self.assertTrue(torch.allclose(K1, K2, atol=1e-4))
        self.assertTrue(torch.allclose(L1, L2, atol=1e-4))

    def test_low_rank_approximation(self):
        cov_sp1 = SpatialCovKernel(
            self.coords, approx_rank=800, centering=True, standardize_cov=False
        )
        cov_sp2 = SpatialCovKernel(
            self.coords, approx_rank=100, centering=True, standardize_cov=False
        )
        self.assertTrue(cov_sp1.rank() == 800)
        self.assertTrue(cov_sp2.rank() == 100)

        L1 = cov_sp1.eigenvalues()
        L2 = cov_sp2.eigenvalues()
        self.assertEqual(len(L2), 100)
        self.assertTrue((L1[:100] - L2).abs().mean() < 0.1)

    def test_kernel_realization(self):
        cov_sp1 = SpatialCovKernel(
            self.coords, approx_rank=None, centering=True, standardize_cov=False
        )
        cov_sp2 = SpatialCovKernel(
            self.coords, approx_rank=100, centering=True, standardize_cov=False
        )

        self.assertTrue(
            (cov_sp1.realization() - cov_sp2.realization()).abs().mean() < 0.1
        )


if __name__ == "__main__":
    unittest.main()



from splisosm.hyptest_np import SplisosmNP
from splisosm.hyptest_glmm import SplisosmGLMM
from splisosm.hyptest_fft import SplisosmFFT

__all__ = ["SplisosmNP", "SplisosmGLMM", "SplisosmFFT", "__version__"]

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("splisosm")
except PackageNotFoundError:
    __version__ = "unknown"



import unittest
import torch

from splisosm.simulation import (
    _sample_multinom_sp_single_gene,
    simulate_isoform_counts_single_gene,
    simulate_isoform_counts,
)


class TestSimulation(unittest.TestCase):
    def test_sample_multinom_sp_single_gene(self):
        # simulate a gene with 3 isoforms in 2 spots
        iso_ratio_expected = torch.tensor([[0.5, 0.3, 0.2], [0.4, 0.4, 0.2]])
        total_counts_expected = 100
        counts = _sample_multinom_sp_single_gene(
            iso_ratio_expected, total_counts_expected
        )

        self.assertEqual(counts.shape, (2, 3))

    def test_simulate_isoform_counts_single_gene(self):
        grid_size = (10, 10)
        n_isos = 3
        total_counts_expected = 100
        var_sp = 0.5
        var_nsp = 0.5
        rho = 0.9

        data, params = simulate_isoform_counts_single_gene(
            grid_size=grid_size,
            n_isos=n_isos,
            total_counts_expected=total_counts_expected,
            var_sp=var_sp,
            var_nsp=var_nsp,
            rho=rho,
            return_params=True,
        )

        self.assertEqual(data["counts"].shape, (grid_size[0] * grid_size[1], n_isos))
        self.assertEqual(data["coords"].shape, (grid_size[0] * grid_size[1], 2))
        self.assertEqual(params["grid_size"], grid_size)
        self.assertEqual(params["n_isos"], n_isos)

    def test_simulate_isoform_counts(self):
        n_genes = 5
        grid_size = (10, 10)
        n_isos = 3
        total_counts_expected = 100
        var_sp = 0.5
        var_nsp = 0.5
        rho = 0.9

        data, params = simulate_isoform_counts(
            n_genes=n_genes,
            grid_size=grid_size,
            n_isos=n_isos,
            total_counts_expected=total_counts_expected,
            var_sp=var_sp,
            var_nsp=var_nsp,
            rho=rho,
            return_params=True,
        )

        self.assertEqual(
            data["counts"].shape, (n_genes, grid_size[0] * grid_size[1], n_isos)
        )
        self.assertEqual(data["coords"].shape, (grid_size[0] * grid_size[1], 2))
        self.assertEqual(params["grid_size"], grid_size)
        self.assertEqual(params["n_isos"], n_isos)
        self.assertEqual(params["n_spots"], grid_size[0] * grid_size[1])


if __name__ == "__main__":
    unittest.main()



import unittest
import torch
import numpy as np
from splisosm.dataset import UngroupedIsoDataset, GroupedIsoDataset, IsoDataset


class TestDatasets(unittest.TestCase):

    def test_ungrouped_iso_dataset(self):
        data = [torch.rand(10, 3), torch.rand(10, 4)]
        gene_names = ["gene1", "gene2"]
        dataset = UngroupedIsoDataset(data, gene_names)

        self.assertEqual(len(dataset), 2)
        self.assertEqual(dataset[0]["n_isos"], 3)
        self.assertEqual(dataset[1]["n_isos"], 4)
        self.assertEqual(dataset[0]["gene_name"], "gene1")
        self.assertEqual(dataset[1]["gene_name"], "gene2")

    def test_grouped_iso_dataset(self):
        data = torch.rand(2, 10, 3)
        gene_names = ["gene1", "gene2"]
        dataset = GroupedIsoDataset(data, gene_names)

        self.assertEqual(len(dataset), 2)
        self.assertEqual(dataset[0]["n_isos"], 3)
        self.assertEqual(dataset[0]["gene_name"], "gene1")
        self.assertEqual(dataset[1]["gene_name"], "gene2")

    def test_iso_dataset_ungrouped(self):
        data = [np.random.rand(10, 3), np.random.rand(10, 4)]
        gene_names = ["gene1", "gene2"]
        dataset = IsoDataset(data, gene_names, group_gene_by_n_iso=False)

        self.assertEqual(len(dataset.datasets), 1)
        self.assertEqual(len(dataset.datasets[0]), 2)
        self.assertEqual(dataset.datasets[0][0]["n_isos"], 3)
        self.assertEqual(dataset.datasets[0][1]["n_isos"], 4)

    def test_iso_dataset_grouped(self):
        data = [np.random.rand(10, 3), np.random.rand(10, 3), np.random.rand(10, 4)]
        gene_names = ["gene1", "gene2", "gene3"]
        dataset = IsoDataset(data, gene_names, group_gene_by_n_iso=True)

        self.assertEqual(len(dataset.datasets), 2)
        self.assertEqual(len(dataset.datasets[0]), 2)  # Two genes with 3 isoforms
        self.assertEqual(len(dataset.datasets[1]), 1)  # One gene with 4 isoforms

    def test_iso_dataset_dataloader(self):
        data = [np.random.rand(10, 3), np.random.rand(10, 4)]
        gene_names = ["gene1", "gene2"]
        dataset = IsoDataset(data, gene_names, group_gene_by_n_iso=False)
        dataloader = dataset.get_dataloader(batch_size=1)

        batches = list(dataloader)
        self.assertEqual(len(batches), 2)
        self.assertEqual(batches[0]["n_isos"].item(), 3)
        self.assertEqual(batches[1]["n_isos"].item(), 4)


if __name__ == "__main__":
    unittest.main()



"""Likelihood functions and pvalue-approximation utilities."""

import numpy as np
from typing import Optional
from numpy.typing import ArrayLike
from scipy.stats import ncx2
import torch

__all__ = [
    "log_prob_fastmult",
    "log_prob_fastmult_batched",
    "log_prob_fastdm",
    "log_prob_fastmvn",
    "log_prob_fastmvn_batched",
    "liu_sf",
]

try:
    from pyro.distributions import Multinomial, DirichletMultinomial, MultivariateNormal
except ImportError:
    from torch.distributions import Multinomial, MultivariateNormal

    DirichletMultinomial = None  # Placeholder if DirichletMultinomial is not available

_DELTA = 1e-10


def log_prob_mult(probs: torch.Tensor, counts: torch.Tensor) -> torch.Tensor:
    """Compute multinomial log-likelihood for one or multiple spots.

    Parameters
    ----------
    probs
        Isoform ratio tensor of shape (n_isos,) or (n_isos, n_spots).
    counts
        Isoform count tensor with the same shape as `probs`.

    Returns
    -------
    torch.Tensor
        Scalar log-likelihood.

    Raises
    ------
    ValueError
        If `probs` and `counts` shapes do not match.
    """
    if probs.shape != counts.shape:
        raise ValueError("`probs` and `counts` must have the same shape.")

    if probs.dim() == 1:  # only one sample (spot)
        log_prob = Multinomial(
            total_count=counts.sum().int().item(), probs=probs
        ).log_prob(counts)
    else:
        log_prob = 0
        total_counts = counts.sum(dim=0).int()  # vector of length n_spots
        for i, total_counts_s in enumerate(total_counts):
            # iterate over samples (spots)
            m = Multinomial(total_count=total_counts_s.item(), probs=probs[:, i])
            log_prob += m.log_prob(counts[:, i])

    return log_prob


def log_prob_fastmult(
    probs: torch.Tensor,
    counts: torch.Tensor,
    mask: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Compute optimized multinomial log-likelihood.

    Parameters
    ----------
    probs
        Isoform ratio tensor of shape (n_isos, n_spots).
    counts
        Isoform count tensor of shape (n_isos, n_spots).
    mask
        Spot mask of shape (n_spots,). Entries with value ``1``
        are treated as masked and are ignored when computing likelihood.

    Returns
    -------
    torch.Tensor
        Scalar log-likelihood.
    """
    n_total = counts.sum(0)  # n_spots
    if mask is not None:
        mask = (1 - mask.int()).bool()
        probs = probs[mask.expand_as(probs)]
        counts = counts[mask.expand_as(counts)]
        n_total = n_total[mask]
    log_prob = (
        (torch.log(probs) * counts).sum()
        + torch.lgamma(n_total + 1).sum()
        - torch.lgamma(counts + 1).sum()
    )

    return log_prob


def log_prob_fastmult_batched(
    probs: torch.Tensor,
    counts: torch.Tensor,
    mask: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Compute batched optimized multinomial log-likelihood.

    Parameters
    ----------
    probs
        Batched isoform ratio tensor of shape (batch_size, n_isos, n_spots).
    counts
        Batched isoform count tensor of shape (batch_size, n_isos, n_spots).
    mask
        Mask tensor of shape (batch_size, n_spots). Entries with value ``1``
        are treated as masked and are ignored when computing likelihood.

    Returns
    -------
    torch.Tensor
        Log-likelihood values of shape (batch_size,).
    """
    batch_size, num_isoforms, num_spots = probs.shape
    if mask is not None:
        mask = (1 - mask.int()).unsqueeze(1)
    else:
        mask = torch.ones(
            batch_size, 1, num_spots, device=probs.device, dtype=probs.dtype
        )
    log_prob = (
        probs.log().mul(counts).mul(mask).sum(dim=[1, 2])
        + counts.sum(dim=1, keepdim=True).add(1).lgamma().mul(mask).sum(dim=[1, 2])
        - counts.add(1).lgamma().mul(mask).sum(dim=[1, 2])
    )
    return log_prob


def log_prob_dm(concentration: torch.Tensor, counts: torch.Tensor) -> torch.Tensor:
    """Compute Dirichlet-multinomial log-likelihood.

    Parameters
    ----------
    concentration
        Isoform concentration tensor of shape (n_isos,) or (n_isos, n_spots).
    counts
        Isoform count tensor with the same shape as `concentration`.

    Returns
    -------
    torch.Tensor
        Scalar log-likelihood.

    Raises
    ------
    ValueError
        If `concentration` and `counts` shapes do not match.
    ImportError
        If `pyro` is not installed and Dirichlet-multinomial likelihood is
        requested.
    """
    if DirichletMultinomial is None:
        raise ImportError(
            "DirichletMultinomial requires `pyro-ppl`. Install `pyro-ppl` to use log_prob_dm()."
        )
    if concentration.shape != counts.shape:
        raise ValueError("`concentration` and `counts` must have the same shape.")

    if concentration.dim() == 1:  # only one sample (spot)
        log_prob = DirichletMultinomial(concentration, counts.sum()).log_prob(counts)
    else:
        log_prob = 0
        total_counts = counts.sum(dim=0)  # vector of length n_spots
        for i, total_counts_s in enumerate(total_counts):
            # iterate over samples (spots)
            dm = DirichletMultinomial(concentration[:, i], total_counts_s)
            log_prob += dm.log_prob(counts[:, i])

    return log_prob


def log_prob_fastdm(
    concentration: torch.Tensor,
    counts: torch.Tensor,
    mask: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Compute optimized Dirichlet-multinomial log-likelihood.

    Parameters
    ----------
    concentration
        Isoform concentration tensor of shape (n_isos, n_spots).
    counts
        Isoform count tensor of shape (n_isos, n_spots).
    mask
        Spot mask of shape (n_spots,). Entries with value ``1``
        are treated as masked and are ignored when computing likelihood.

    Returns
    -------
    torch.Tensor
        Scalar log-likelihood.
    """
    n_total_conc = concentration.sum(0)  # n_spots
    n_total_counts = counts.sum(0)  # n_spots

    if mask is not None:
        mask = (1 - mask.int()).bool()
        concentration = concentration[:, mask]
        counts = counts[:, mask]
        n_total_conc = n_total_conc[mask]
        n_total_counts = n_total_counts[mask]

    log_prob = (
        torch.lgamma(concentration + counts).sum()
        - torch.lgamma(concentration).sum()
        - torch.lgamma(counts + 1).sum()
        + torch.lgamma(n_total_conc).sum()
        + torch.lgamma(n_total_counts + 1).sum()
        - torch.lgamma(n_total_conc + n_total_counts).sum()
    )

    return log_prob


def log_prob_mvn(
    locs: torch.Tensor,
    covs: torch.Tensor,
    data: torch.Tensor,
) -> torch.Tensor:
    """Compute multivariate normal log-likelihood.

    When multiple classes (e.g., isoforms) are present, they are treated as independent,
    and the log-likelihood is computed as the sum of log-likelihoods for each class.

    Parameters
    ----------
    locs
        Mean tensor of shape (n_spots,) or (n_classes, n_spots).
    covs
        Covariance tensor of shape (n_spots, n_spots) or
        (n_classes, n_spots, n_spots).
    data
        Observations of shape (n_spots,) or (n_classes, n_spots).

    Returns
    -------
    torch.Tensor
        Scalar log-likelihood.

    Raises
    ------
    ValueError
        If batched dimensions are inconsistent.
    """
    if data.dim() == 1:  # only one sample (isoform)
        mvn = MultivariateNormal(locs, covariance_matrix=covs)
        log_prob = mvn.log_prob(data)
    else:
        n_isos = data.shape[0]
        if len(locs) != n_isos or len(covs) != n_isos:
            raise ValueError(
                "For batched input, `locs` and `covs` must match the number of isoforms in `data`."
            )

        log_prob = 0
        for mu_i, cov_i, gamma_i in zip(locs, covs, data):
            # iterate over samples (isoforms)
            mvn = MultivariateNormal(mu_i, covariance_matrix=cov_i)
            log_prob += mvn.log_prob(gamma_i)

    return log_prob


def log_prob_fastmvn(
    locs: torch.Tensor,
    cov_eigvals: torch.Tensor,
    cov_eigvecs: torch.Tensor,
    data: torch.Tensor,
    mask: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Compute multivariate normal log-likelihood with eigendecomposition.

    Classes along the first axis are treated as independent,
    and the log-likelihood is computed as the sum of log-likelihoods for each class.

    Parameters
    ----------
    locs
        Mean tensor of shape (n_classes, n_spots).
    cov_eigvals
        Covariance eigenvalues of shape (n_classes, n_spots) or (1, n_spots).
    cov_eigvecs
        Covariance eigenvectors of shape (n_classes, n_spots, n_spots) or (1, n_spots, n_spots).
    data
        Observation tensor of shape (n_classes, n_spots).
    mask
        Spot mask of shape (n_spots,). Entries with value ``1``
        are treated as masked and are ignored when computing likelihood.

    Returns
    -------
    torch.Tensor
        Scalar log-likelihood.

    Raises
    ------
    ValueError
        If covariance eigendecomposition shapes are incompatible with `data`.
    """
    n_classes, n_spots = data.shape
    if cov_eigvals.shape not in {(n_classes, n_spots), (1, n_spots)}:
        raise ValueError(
            "`cov_eigvals` must have shape (n_classes, n_spots) or (1, n_spots)."
        )
    if cov_eigvecs.shape not in {(n_classes, n_spots, n_spots), (1, n_spots, n_spots)}:
        raise ValueError(
            "`cov_eigvecs` must have shape (n_classes, n_spots, n_spots) or (1, n_spots, n_spots)."
        )

    if mask is not None:
        mask = (1 - mask.int()).unsqueeze(0)
        locs = locs * mask
        data = data * mask
        n_spots = mask.sum()

    # calculate (x - mu).T @ (VLV.T)^-1 @ (x - mu)
    res = ((data - locs).unsqueeze(1) @ cov_eigvecs).squeeze(1)  # n_classes x n_spots
    quad_gamma = (res**2 / cov_eigvals).sum()

    # calculate log_det
    log_det_cov = torch.log(cov_eigvals).sum() * (
        n_classes if cov_eigvals.shape[0] == 1 else 1
    )

    log_prob = -0.5 * (
        n_spots * n_classes * np.log(2 * np.pi) + log_det_cov + quad_gamma
    )

    return log_prob


def log_prob_fastmvn_batched(
    locs: torch.Tensor,
    cov_eigvals: torch.Tensor,
    cov_eigvecs: torch.Tensor,
    data: torch.Tensor,
    mask: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Compute batched multivariate normal log-likelihood with eigendecomposition.

    Classes along the second axis are treated as independent,
    and the log-likelihood is computed as the sum of log-likelihoods for each class.

    Parameters
    ----------
    locs
        Batched mean tensor of shape (batch_size, n_classes, n_spots).
    cov_eigvals
        Batched eigenvalue tensor of shape (batch_size, n_classes, n_spots).
    cov_eigvecs
        Batched eigenvector tensor of shape (batch_size, n_classes, n_spots, n_spots).
    data
        Batched observation of shape (batch_size, n_classes, n_spots).
    mask : torch.Tensor, optional
        Batched mask tensor of shape (batch_size, n_spots). Entries with value ``1``
        are treated as masked and are ignored when computing likelihood.

    Returns
    -------
    torch.Tensor
        Log-likelihood values of shape (batch_size,).

    Raises
    ------
    ValueError
        If covariance decomposition shapes are incompatible with input sizes.
    """
    batch_size, num_classes, num_spots = data.shape
    if mask is not None:
        mask = (1 - mask.int()).unsqueeze(1)
        num_spots_non_mask = mask.sum(dim=[1, 2])
    else:
        num_spots_non_mask = num_spots
        mask = torch.ones(
            batch_size, 1, num_spots, device=data.device, dtype=data.dtype
        )

    if cov_eigvals.shape[1:] != (num_classes, num_spots):
        raise ValueError(f"Invalid shape of cov_eigvals: {cov_eigvals.shape}")
    if cov_eigvecs.shape[1:] != (num_classes, num_spots, num_spots):
        raise ValueError(f"Invalid shape of cov_eigvecs: {cov_eigvecs.shape}")

    # calculate (x - mu).T @ (VLV.T)^-1 @ (x - mu)
    # batch_size x n_classes x n_spots
    # res = ((data - locs).mul(mask).unsqueeze(2) @ cov_eigvecs).squeeze(2)
    res = torch.einsum("bis,bist->bit", (data - locs).mul(mask), cov_eigvecs)
    quad_gamma = (res**2 / cov_eigvals).sum(dim=[1, 2])

    # calculate log_det
    log_det_cov = cov_eigvals.log().sum(dim=[1, 2])
    log_2pi = torch.log(torch.tensor(2.0 * np.pi, device=data.device, dtype=data.dtype))
    log_prob = -0.5 * (
        num_spots_non_mask * num_classes * log_2pi + log_det_cov + quad_gamma
    )

    return log_prob


def liu_sf(
    t: ArrayLike,
    lambs: ArrayLike,
    dofs: Optional[ArrayLike] = None,
    deltas: Optional[ArrayLike] = None,
    kurtosis: bool = False,
) -> np.ndarray:
    """Compute p-values for weighted sums of chi-squared variables using the Liu approximation.

    Let :math:`X = \\sum_{i=1}^{d} \\lambda_i * \\chi^2(h_i, \\delta_i)` be a linear combination of
    `d` chi-squared random variables, each with degree of freedom :math:`h_i` and noncentrality
    :math:`\\delta_i`, this function approximates :math:``Pr(X > t)`` using the
    Liu moment-matching approach :cite:`liu2009new`.

    Parameters
    ----------
    t
        Shape (n,), observed test statistic.
    lambs
        Shape (d,), weights of each chi-squared component.
    dofs
        Shape (d,), degrees of freedom for each component. Defaults to all ones.
    deltas
        Shape (d,), noncentrality parameters for each component. Defaults to all zeros.
    kurtosis
        If True, uses kurtosis matching proposed in :cite:`lee2012optimal`; otherwise uses the original
        skewness matching :cite:`liu2009new`.

    Returns
    -------
    numpy.ndarray
        P-values of shape (n,) computed as ``Pr(X > t)``.

    Notes
    -----
    From https://github.com/limix/chiscore/blob/master/chiscore/_liu.py
    """
    if dofs is None:
        dofs = np.ones_like(lambs)
    if deltas is None:
        deltas = np.zeros_like(lambs)

    t = np.asarray(t, float)
    lambs = np.asarray(lambs, float)
    dofs = np.asarray(dofs, float)
    deltas = np.asarray(deltas, float)

    lambs = {i: lambs**i for i in range(1, 5)}

    c = {
        i: np.sum(lambs[i] * dofs) + i * np.sum(lambs[i] * deltas) for i in range(1, 5)
    }

    s1 = c[3] / (np.sqrt(c[2]) ** 3 + _DELTA)
    s2 = c[4] / (c[2] ** 2 + _DELTA)

    s12 = s1**2
    if s12 > s2:
        a = 1 / (s1 - np.sqrt(s12 - s2))
        delta_x = s1 * a**3 - a**2
        dof_x = a**2 - 2 * delta_x
    else:
        delta_x = 0
        if kurtosis:
            a = 1 / np.sqrt(s2)
            dof_x = 1 / s2
        else:
            a = 1 / (s1 + _DELTA)
            dof_x = 1 / (s12 + _DELTA)

    mu_q = c[1]
    sigma_q = np.sqrt(2 * c[2])

    mu_x = dof_x + delta_x
    sigma_x = np.sqrt(2 * (dof_x + 2 * delta_x))

    t_star = (t - mu_q) / (sigma_q + _DELTA)
    tfinal = t_star * sigma_x + mu_x

    q = ncx2.sf(tfinal, dof_x, np.maximum(delta_x, 1e-9))

    return q



"""Unit tests for splisosm.kernel_gpr."""

import unittest
import warnings

import numpy as np
import torch

from splisosm.kernel_gpr import (
    DenseKernelOp,
    FFTKernelOp,
    FFTKernelGPR,
    SpatialKernelOp,
    SklearnKernelGPR,
    GPyTorchKernelGPR,
    make_kernel_gpr,
    linear_hsic_test,
    _build_rbf_kernel,
    _DEFAULT_GPR_CONFIGS,
    _kernel_residuals_from_eigdecomp,
)


def _make_coords(n: int = 50, d: int = 2, seed: int = 0) -> torch.Tensor:
    torch.manual_seed(seed)
    coords = torch.randn(n, d)
    return (coords - coords.mean(0)) / coords.std(0)


def _make_rbf_kernel(
    n: int = 50, constant_value: float = 1.0, length_scale: float = 1.0
):
    coords = _make_coords(n)
    return _build_rbf_kernel(coords, constant_value, length_scale)


class TestBuildRbfKernel(unittest.TestCase):
    def test_shape(self):
        K = _make_rbf_kernel(n=30)
        self.assertEqual(K.shape, (30, 30))

    def test_symmetric_psd(self):
        K = _make_rbf_kernel(n=40)
        self.assertTrue(torch.allclose(K, K.T, atol=1e-6))
        evals = torch.linalg.eigvalsh(K)
        self.assertTrue((evals >= -1e-6).all(), "Kernel has negative eigenvalues")

    def test_diagonal_equals_constant_value(self):
        K = _make_rbf_kernel(n=20, constant_value=2.5)
        self.assertTrue(torch.allclose(K.diag(), torch.full((20,), 2.5), atol=1e-5))


class TestDenseKernelOp(unittest.TestCase):
    def setUp(self):
        self.n = 30
        self.K = _make_rbf_kernel(self.n)
        self.op = DenseKernelOp(self.K)

    def test_is_spatial_kernel_op(self):
        self.assertIsInstance(self.op, SpatialKernelOp)

    def test_n(self):
        self.assertEqual(self.op.n, self.n)

    def test_matvec_shape(self):
        torch.manual_seed(0)
        v = torch.randn(self.n, 5)
        out = self.op.matvec(v)
        self.assertEqual(out.shape, v.shape)

    def test_solve_1d_and_2d(self):
        torch.manual_seed(0)
        v1 = torch.randn(self.n)
        v2 = torch.randn(self.n, 4)
        u1 = self.op.solve(v1, epsilon=0.1)
        u2 = self.op.solve(v2, epsilon=0.1)
        self.assertEqual(u1.shape, v1.shape)
        self.assertEqual(u2.shape, v2.shape)

    def test_residuals_consistent(self):
        torch.manual_seed(0)
        v = torch.randn(self.n, 3)
        epsilon = 0.05
        res1 = self.op.residuals(v, epsilon)
        res2 = epsilon * self.op.solve(v, epsilon)
        self.assertTrue(torch.allclose(res1, res2, atol=1e-5))

    def test_eigenvalues_descending(self):
        evals = self.op.eigenvalues()
        self.assertEqual(len(evals), self.n)
        self.assertTrue((evals[:-1] >= evals[1:] - 1e-6).all())

    def test_eigenvalues_k(self):
        evals_k = self.op.eigenvalues(k=5)
        evals_all = self.op.eigenvalues()
        self.assertTrue(torch.allclose(evals_k, evals_all[:5], atol=1e-6))

    def test_trace(self):
        tr = self.op.trace()
        self.assertAlmostEqual(tr, float(self.K.diag().sum()), places=3)

    def test_eigh(self):
        vals, vecs = self.op.eigh()
        self.assertEqual(vals.shape, (self.n,))
        self.assertEqual(vecs.shape, (self.n, self.n))
        # Columns should be orthonormal
        self.assertTrue(torch.allclose(vecs.T @ vecs, torch.eye(self.n), atol=1e-5))

    def test_chol_caching(self):
        """Cholesky is cached and reused for the same epsilon."""
        torch.manual_seed(0)
        v = torch.randn(self.n)
        epsilon = 0.1
        u1 = self.op.solve(v, epsilon)
        chol1 = self.op._chol
        u2 = self.op.solve(v, epsilon)
        chol2 = self.op._chol
        self.assertIs(chol1, chol2)  # same object (cached)
        self.assertTrue(torch.allclose(u1, u2))


class TestLinearHsicTest(unittest.TestCase):
    def _make_data(self, n=80, nx=2, ny=3, seed=0):
        torch.manual_seed(seed)
        X = torch.randn(n, nx)
        Y = torch.randn(n, ny)
        return X, Y

    def test_returns_floats(self):
        X, Y = self._make_data()
        hsic, pval = linear_hsic_test(X, Y)
        self.assertIsInstance(hsic, float)
        self.assertIsInstance(pval, float)

    def test_pval_in_01(self):
        X, Y = self._make_data()
        _, pval = linear_hsic_test(X, Y)
        self.assertGreaterEqual(pval, 0.0)
        self.assertLessEqual(pval, 1.0)

    def test_independent_data_high_pval(self):
        """Uncorrelated X, Y should give non-significant p-values on average."""
        torch.manual_seed(42)
        pvals = []
        for _ in range(50):
            X = torch.randn(100, 2)
            Y = torch.randn(100, 3)
            _, pval = linear_hsic_test(X, Y)
            pvals.append(pval)
        # Under independence, p-values uniform(0,1) → mean ~0.5
        self.assertGreater(np.mean(pvals), 0.2)

    def test_dependent_data_low_pval(self):
        """Highly correlated X, Y should give small p-values."""
        torch.manual_seed(0)
        X = torch.randn(100, 2)
        Y = X + 0.01 * torch.randn(100, 2)  # nearly identical
        _, pval = linear_hsic_test(X, Y)
        self.assertLess(pval, 0.05)

    def test_nan_handling(self):
        """NaN rows should be silently removed."""
        torch.manual_seed(0)
        X = torch.randn(80, 2)
        Y = torch.randn(80, 3)
        X[5] = float("nan")
        Y[10, 1] = float("nan")
        hsic, pval = linear_hsic_test(X, Y)
        self.assertTrue(np.isfinite(hsic))
        self.assertTrue(np.isfinite(pval))


class TestKernelResidualFromEigdecomp(unittest.TestCase):
    def test_output_shape(self):
        n, m = 40, 5
        K = _make_rbf_kernel(n)
        evals, evecs = torch.linalg.eigh(K)
        Y = torch.randn(n, m)
        res = _kernel_residuals_from_eigdecomp(evecs, evals, Y)
        self.assertEqual(res.shape, (n, m))

    def test_residuals_smaller_than_input(self):
        n = 50
        K = _make_rbf_kernel(n, constant_value=2.0)
        evals, evecs = torch.linalg.eigh(K)
        torch.manual_seed(1)
        Y = torch.randn(n, 4)
        res = _kernel_residuals_from_eigdecomp(evecs, evals, Y)
        self.assertLess(float(res.norm()), float(Y.norm()))


class TestSklearnKernelGPR(unittest.TestCase):
    def setUp(self):
        self.n = 60
        self.coords = _make_coords(self.n)
        torch.manual_seed(42)
        self.Y = torch.randn(self.n, 3)

    def _make_gpr_fixed(self):
        return SklearnKernelGPR(
            constant_value=1e-3,
            constant_value_bounds="fixed",
            length_scale=1.0,
            length_scale_bounds="fixed",
        )

    def _make_gpr_free(self):
        return SklearnKernelGPR(
            constant_value=1.0,
            constant_value_bounds=(1e-3, 1e3),
            length_scale=1.0,
            length_scale_bounds="fixed",
        )

    def test_from_config_default(self):
        # Both covariate and isoform configs now use free (calibrated) bounds.
        for key in ("covariate", "isoform"):
            gpr = SklearnKernelGPR.from_config(_DEFAULT_GPR_CONFIGS[key])
            self.assertFalse(
                gpr.signal_bounds_fixed,
                f"_DEFAULT_GPR_CONFIGS['{key}'] should have free constant_value_bounds",
            )

    def test_signal_bounds_fixed(self):
        gpr_f = self._make_gpr_fixed()
        gpr_m = self._make_gpr_free()
        self.assertTrue(gpr_f.signal_bounds_fixed)
        self.assertFalse(gpr_m.signal_bounds_fixed)

    def test_fit_residuals_shape(self):
        gpr = self._make_gpr_fixed()
        res = gpr.fit_residuals(self.coords, self.Y)
        self.assertEqual(res.shape, self.Y.shape)

    def test_fit_residuals_nan_handling(self):
        gpr = self._make_gpr_free()
        Y_nan = self.Y.clone()
        Y_nan[5] = float("nan")
        res = gpr.fit_residuals(self.coords, Y_nan)
        self.assertEqual(res.shape, self.Y.shape)
        self.assertTrue(torch.isnan(res[5]).all())
        self.assertFalse(torch.isnan(res[6]).any())

    def test_shared_eigendecomp_fast_path(self):
        gpr = self._make_gpr_fixed()
        # After precompute, fit_residuals should use the fast path
        gpr.precompute_shared_kernel(self.coords)
        self.assertIsNotNone(gpr._shared_eigvals)
        res = gpr.fit_residuals(self.coords, self.Y)
        self.assertEqual(res.shape, self.Y.shape)

    def test_precompute_warns_when_not_fixed(self):
        gpr = self._make_gpr_free()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            gpr.precompute_shared_kernel(self.coords)
            self.assertTrue(any("no effect" in str(x.message) for x in w))

    def test_fit_residuals_batch_amortizes(self):
        gpr = self._make_gpr_fixed()
        Y_list = [torch.randn(self.n, 3) for _ in range(4)]
        res_list = gpr.fit_residuals_batch(self.coords, Y_list)
        self.assertEqual(len(res_list), 4)
        for res in res_list:
            self.assertEqual(res.shape, (self.n, 3))
        # After batch, eigendecomp should be precomputed
        self.assertIsNotNone(gpr._shared_eigvals)

    def test_fast_path_shape_guard_with_nan(self):
        """Fast path should not be used when NaN rows cause coord subset."""
        gpr = self._make_gpr_fixed()
        gpr.precompute_shared_kernel(self.coords)
        Y_nan = self.Y.clone()
        Y_nan[0:5] = float("nan")
        # Should not raise (falls back to sklearn GPR for the NaN-filtered subset)
        res = gpr.fit_residuals(self.coords, Y_nan)
        self.assertEqual(res.shape, self.Y.shape)
        self.assertTrue(torch.isnan(res[:5]).all())

    def test_get_kernel_op(self):
        from splisosm.kernel_gpr import DenseKernelOp

        gpr = self._make_gpr_fixed()
        op = gpr.get_kernel_op(self.coords)
        self.assertIsInstance(op, DenseKernelOp)


class TestGPyTorchKernelGPRImportGuard(unittest.TestCase):
    def test_raises_without_gpytorch(self):
        """GPyTorchKernelGPR raises ImportError if gpytorch not installed."""
        import sys

        # If gpytorch is available, skip; if not, ensure ImportError on instantiation
        if "gpytorch" in sys.modules:
            self.skipTest("gpytorch is available; skipping import guard test.")
        try:
            GPyTorchKernelGPR()
        except ImportError:
            pass  # expected
        except Exception as e:
            # Some other error is also acceptable if gpytorch isn't present
            self.assertIsInstance(e, (ImportError, RuntimeError))

    def test_fit_residuals_shape(self):
        """fit_residuals returns correct shape for multi-output Y."""
        try:
            gpr = GPyTorchKernelGPR(n_iter=5)
        except ImportError:
            self.skipTest("gpytorch not installed.")
        n, m = 20, 3
        coords = torch.randn(n, 2)
        Y = torch.randn(n, m)
        res = gpr.fit_residuals(coords, Y)
        self.assertEqual(res.shape, (n, m))

    def test_fit_residuals_nan_rows_preserved(self):
        """NaN rows in Y are reinserted as NaN in the residuals."""
        try:
            gpr = GPyTorchKernelGPR(n_iter=5)
        except ImportError:
            self.skipTest("gpytorch not installed.")
        n, m = 20, 2
        coords = torch.randn(n, 2)
        Y = torch.randn(n, m)
        Y[3, :] = float("nan")
        res = gpr.fit_residuals(coords, Y)
        self.assertEqual(res.shape, (n, m))
        self.assertTrue(torch.isnan(res[3]).all())
        self.assertFalse(torch.isnan(res[0]).any())

    def test_fit_residuals_shrinks_signal(self):
        """Residuals have smaller variance than pure spatial signal."""
        try:
            gpr = GPyTorchKernelGPR(
                constant_value=1.0,
                constant_value_bounds="fixed",
                length_scale=1.0,
                length_scale_bounds="fixed",
                n_iter=20,
            )
        except ImportError:
            self.skipTest("gpytorch not installed.")
        torch.manual_seed(0)
        n = 30
        t = torch.linspace(0, 2 * np.pi, n).unsqueeze(1)
        coords = torch.cat([t, torch.zeros_like(t)], dim=1)
        Y = torch.sin(t).expand(n, 3) + 0.05 * torch.randn(n, 3)
        res = gpr.fit_residuals(coords, Y)
        self.assertLess(res.std().item(), Y.std().item())

    def test_n_inducing_accepted(self):
        """n_inducing > 0 is now accepted (FITC sparse GP path)."""
        try:
            gpr = GPyTorchKernelGPR(n_inducing=50)
            self.assertIsNotNone(gpr)
        except ImportError:
            self.skipTest("gpytorch not installed.")


class TestMakeKernelGPR(unittest.TestCase):
    def test_sklearn_backend(self):
        gpr = make_kernel_gpr(
            "sklearn", constant_value=0.5, constant_value_bounds="fixed"
        )
        self.assertIsInstance(gpr, SklearnKernelGPR)
        self.assertTrue(gpr.signal_bounds_fixed)

    def test_invalid_backend(self):
        with self.assertRaises(ValueError):
            make_kernel_gpr("nonexistent_backend")

    def test_default_configs_structure(self):
        self.assertIn("covariate", _DEFAULT_GPR_CONFIGS)
        self.assertIn("isoform", _DEFAULT_GPR_CONFIGS)
        for key in (
            "constant_value",
            "constant_value_bounds",
            "length_scale",
            "length_scale_bounds",
        ):
            self.assertIn(key, _DEFAULT_GPR_CONFIGS["covariate"])
            self.assertIn(key, _DEFAULT_GPR_CONFIGS["isoform"])


def _make_regular_grid_coords(ny: int = 15, nx: int = 20) -> torch.Tensor:
    """Return z-score normalized coords for a regular (ny, nx) grid."""
    y_idx = torch.arange(ny, dtype=torch.float32)
    x_idx = torch.arange(nx, dtype=torch.float32)
    yy, xx = torch.meshgrid(y_idx, x_idx, indexing="ij")
    coords = torch.stack([yy.flatten(), xx.flatten()], dim=1)
    coords = (coords - coords.mean(0)) / coords.std(0)
    return coords


class TestFFTKernelOp(unittest.TestCase):
    """Tests for FFTKernelOp."""

    def setUp(self):
        self.op = FFTKernelOp(
            ny=15,
            nx=20,
            dy=0.1,
            dx=0.1,
            constant_value=1.0,
            length_scale=0.5,
        )

    def test_n(self):
        self.assertEqual(self.op.n, 300)
        self.assertEqual(self.op.ny, 15)
        self.assertEqual(self.op.nx, 20)

    def test_eigenvalues_shape_and_sorted(self):
        evals = self.op.eigenvalues()
        self.assertEqual(evals.shape, (300,))
        self.assertGreaterEqual(float(evals[0]), float(evals[-1]))

    def test_eigenvalues_nonneg(self):
        evals = self.op.eigenvalues()
        self.assertTrue(np.all(evals >= 0), "FFT eigenvalues should be non-negative")

    def test_trace_equals_n_times_constant_value(self):
        # trace(K) = sum of eigenvalues ≈ n * constant_value for large grid
        # (exact for periodic torus); here we just check it's positive and finite
        t = self.op.trace()
        self.assertGreater(t, 0)
        self.assertTrue(np.isfinite(t))

    def test_matvec_shape_1d(self):
        v = np.random.randn(300)
        kv = self.op.matvec(v)
        self.assertEqual(kv.shape, (300,))

    def test_matvec_shape_2d(self):
        V = np.random.randn(300, 5)
        KV = self.op.matvec(V)
        self.assertEqual(KV.shape, (300, 5))

    def test_residuals_reduce_variance(self):
        """GP residualization should reduce signal variance."""
        np.random.seed(7)
        v = np.random.randn(300)
        r = self.op.residuals(v, epsilon=0.5)
        self.assertLess(float(np.var(r)), float(np.var(v)))

    def test_solve_is_inverse_of_matvec_plus_eps(self):
        """solve((K + eps*I)v) ≈ v."""
        eps = 0.2
        np.random.seed(3)
        v = np.random.randn(300)
        u = self.op.solve(v, eps)
        # Reconstruct: matvec(u) + eps*u should ≈ v
        recon = self.op.matvec(u) + eps * u
        np.testing.assert_allclose(recon, v, atol=1e-6)


class TestFFTKernelGPR(unittest.TestCase):
    """Tests for FFTKernelGPR."""

    def setUp(self):
        self.coords = _make_regular_grid_coords(ny=15, nx=20)
        n = 300
        # Smooth spatial signal + noise.
        # Use frequency 1.5 rad/unit — within the RBF kernel's spectral support on
        # this 15×20 grid (ky=2 at f≈1.94 has negative circulant eigenvalue; f=1.5
        # falls clearly at ky=1 which has eigenvalue ≈35).
        torch.manual_seed(42)
        self.Y_signal = torch.sin(self.coords[:, 0] * 1.5).unsqueeze(
            1
        ) + 0.05 * torch.randn(n, 1)
        self.Y_noise = torch.randn(n, 2)

    def test_fit_residuals_shape(self):
        gpr = FFTKernelGPR(constant_value_bounds="fixed")
        res = gpr.fit_residuals(self.coords, self.Y_signal)
        self.assertEqual(res.shape, self.Y_signal.shape)

    def test_fit_residuals_reduces_spatial_signal(self):
        """Residuals should have substantially less variance than a smooth signal."""
        gpr = FFTKernelGPR(constant_value_bounds=(1e-3, 1e3))
        res = gpr.fit_residuals(self.coords, self.Y_signal)
        self.assertLess(float(res.var()), float(self.Y_signal.var()) * 0.5)

    def test_fit_residuals_noise_unchanged(self):
        """Residuals of pure noise should retain most variance (kernel can't fit noise)."""
        gpr = FFTKernelGPR(constant_value_bounds=(1e-3, 1e3))
        res = gpr.fit_residuals(self.coords, self.Y_noise)
        # Residual variance should be a large fraction of original
        self.assertGreater(float(res.var()), float(self.Y_noise.var()) * 0.3)

    def test_fit_residuals_nan_rows_preserved(self):
        """NaN rows in Y should be preserved as NaN in residuals."""
        Y = self.Y_signal.clone()
        Y[5] = float("nan")
        Y[20] = float("nan")
        gpr = FFTKernelGPR(constant_value_bounds="fixed")
        res = gpr.fit_residuals(self.coords, Y)
        self.assertTrue(torch.isnan(res[5]).all())
        self.assertTrue(torch.isnan(res[20]).all())
        self.assertFalse(torch.isnan(res[0]).any())

    def test_fit_residuals_batch_matches_single(self):
        """fit_residuals_batch should give same result as repeated fit_residuals."""
        gpr = FFTKernelGPR(constant_value_bounds="fixed")
        res_single = [
            gpr.fit_residuals(self.coords, self.Y_signal),
            gpr.fit_residuals(self.coords, self.Y_noise),
        ]
        res_batch = gpr.fit_residuals_batch(self.coords, [self.Y_signal, self.Y_noise])
        for single, batch in zip(res_single, res_batch):
            torch.testing.assert_close(single, batch, atol=1e-5, rtol=1e-4)

    def test_fit_residuals_cube_backward_compat(self):
        """fit_residuals_cube should work and return (array, float)."""
        gpr = FFTKernelGPR(constant_value_bounds="fixed")
        cube = np.random.randn(8, 10)
        res_cube, eps = gpr.fit_residuals_cube(cube, spacing=(0.125, 0.1))
        self.assertEqual(res_cube.shape, (8, 10))
        self.assertGreater(eps, 0)

    def test_irregular_grid_raises(self):
        """fit_residuals should raise ValueError for irregular coordinates."""
        gpr = FFTKernelGPR()
        irregular = self.coords.clone()
        irregular[5, 0] += 0.0137  # break regularity
        with self.assertRaises(ValueError):
            gpr.fit_residuals(irregular, self.Y_signal)

    def test_length_scale_bounds_invalid_raises(self):
        """Invalid length_scale_bounds should raise ValueError."""
        with self.assertRaises(ValueError):
            FFTKernelGPR(length_scale_bounds="auto")
        with self.assertRaises(ValueError):
            FFTKernelGPR(length_scale_bounds=(10.0, 0.1))  # lo >= hi

    def test_length_scale_optimization_free(self):
        """Free length_scale_bounds: completes, correct shape, reduces variance."""
        gpr = FFTKernelGPR(
            constant_value_bounds="fixed",
            length_scale_bounds=(0.1, 10.0),
        )
        res = gpr.fit_residuals(self.coords, self.Y_signal)
        self.assertEqual(res.shape, self.Y_signal.shape)
        self.assertLess(float(res.var()), float(self.Y_signal.var()))

    def test_make_kernel_gpr_fft(self):
        gpr = make_kernel_gpr("fft", **_DEFAULT_GPR_CONFIGS["covariate"])
        self.assertIsInstance(gpr, FFTKernelGPR)
        res = gpr.fit_residuals(self.coords, self.Y_signal)
        self.assertEqual(res.shape, self.Y_signal.shape)

    def test_fft_matches_sklearn_calibration(self):
        """FFT and sklearn backends should give similar calibration on null data."""
        from scipy.stats import kstest

        torch.manual_seed(99)
        np.random.seed(99)
        n = self.coords.shape[0]  # 300 (15x20 grid)
        n_genes = 40

        # Null isoform ratios (no spatial signal)
        Y_list = [torch.rand(n, 3) for _ in range(n_genes)]

        # Residualize a single covariate
        covar_config = _DEFAULT_GPR_CONFIGS["covariate"]
        z = torch.randn(n, 1)

        gpr_sk = SklearnKernelGPR(**covar_config)
        z_sk = gpr_sk.fit_residuals(self.coords, z)

        gpr_fft = FFTKernelGPR(
            constant_value=covar_config["constant_value"],
            constant_value_bounds=covar_config["constant_value_bounds"],
            length_scale=covar_config["length_scale"],
            length_scale_bounds=covar_config["length_scale_bounds"],
        )
        z_fft = gpr_fft.fit_residuals(self.coords, z)

        pvals_sk = [linear_hsic_test(z_sk, Y, centering=True)[1] for Y in Y_list]
        pvals_fft = [linear_hsic_test(z_fft, Y, centering=True)[1] for Y in Y_list]

        # Both should be roughly uniform under null (KS p > 0.05)
        ks_sk = kstest(pvals_sk, "uniform").pvalue
        ks_fft = kstest(pvals_fft, "uniform").pvalue
        self.assertGreater(
            ks_sk, 0.01, f"sklearn p-values not uniform: KS p={ks_sk:.4f}"
        )
        self.assertGreater(ks_fft, 0.01, f"FFT p-values not uniform: KS p={ks_fft:.4f}")

    def test_fft_sklearn_du_agreement(self):
        """FFT and sklearn give highly correlated residuals and p-values on realistic covariates.

        Uses a mixed-frequency covariate (low-freq spatial + high-freq + noise) so that
        neither backend can achieve eps→0 (degenerate interpolation), ensuring a fair
        comparison.  The same calibrated config (case 2: free sigma2, fixed length_scale)
        is used for both backends.
        """
        from scipy.stats import spearmanr

        torch.manual_seed(0)
        n = self.coords.shape[0]  # 300

        # Mixed covariate: low-freq + high-freq + noise (matches realistic covariates)
        z = (
            torch.sin(self.coords[:, 0] * 1.5) * 0.5  # low-frequency spatial
            + torch.sin(self.coords[:, 0] * 8.0) * 0.3  # high-frequency spatial
            + torch.randn(n) * 0.4  # independent noise
        ).unsqueeze(1)
        z = (z - z.mean()) / z.std()

        # Same calibrated config for both backends (case 2: free sigma2, fixed length_scale)
        cfg = dict(
            constant_value=1.0,
            constant_value_bounds=(1e-3, 1e3),
            length_scale=1.0,
            length_scale_bounds="fixed",
        )
        gpr_fft = FFTKernelGPR(**cfg)
        z_fft = gpr_fft.fit_residuals(self.coords, z)

        gpr_sk = SklearnKernelGPR(**cfg)
        z_sk = gpr_sk.fit_residuals(self.coords, z)

        # (1) Residuals should be highly correlated (same GP kernel, different approximation)
        r = float(np.corrcoef(z_fft.squeeze().numpy(), z_sk.squeeze().numpy())[0, 1])
        self.assertGreater(
            r, 0.8, f"Pearson r between FFT and sklearn residuals: {r:.3f}"
        )

        # (2) HSIC p-values for 40 null + 40 signal genes should agree in ranking
        pv_fft, pv_sk = [], []
        for seed in range(40):
            torch.manual_seed(seed + 100)
            Y_null = torch.randn(n, 2)
            _, p_null_fft = linear_hsic_test(z_fft, Y_null, centering=True)
            _, p_null_sk = linear_hsic_test(z_sk, Y_null, centering=True)
            pv_fft.append(p_null_fft)
            pv_sk.append(p_null_sk)
        # Add signal genes (correlated with z — should get low p-values)
        torch.manual_seed(77)
        for _ in range(40):
            Y_sig = 0.7 * z + 0.3 * torch.randn(n, 2)
            _, p_sig_fft = linear_hsic_test(z_fft, Y_sig, centering=True)
            _, p_sig_sk = linear_hsic_test(z_sk, Y_sig, centering=True)
            pv_fft.append(p_sig_fft)
            pv_sk.append(p_sig_sk)

        rho, _ = spearmanr(pv_fft, pv_sk)
        self.assertGreater(
            rho,
            0.8,
            f"Spearman rho between FFT and sklearn DU p-values: {rho:.3f} (expected > 0.8)",
        )


try:
    import gpytorch as _gpytorch_check  # noqa: F401

    _GPYTORCH_AVAILABLE = True
except ImportError:
    _GPYTORCH_AVAILABLE = False


class TestSklearnKernelGPRLargeN(unittest.TestCase):
    """Tests for the large-n approximate path (subset GP + chunked prediction)."""

    def setUp(self):
        torch.manual_seed(7)
        np.random.seed(7)
        self.coords = _make_coords(300)  # 300 points
        signal = torch.sin(self.coords[:, 0] * 2.0)
        self.Y = signal.unsqueeze(1) + 0.1 * torch.randn(300, 1)

    def test_large_n_output_shape(self):
        """Approximate path should return same shape as input."""
        gpr = SklearnKernelGPR(
            constant_value_bounds=(1e-3, 1e3),
            length_scale_bounds="fixed",
            n_inducing=50,
        )
        res = gpr.fit_residuals(self.coords, self.Y)
        self.assertEqual(res.shape, self.Y.shape)

    def test_large_n_reduces_signal(self):
        """Approximate path should substantially reduce spatial signal variance."""
        gpr = SklearnKernelGPR(
            constant_value_bounds=(1e-3, 1e3),
            length_scale_bounds="fixed",
            n_inducing=50,
        )
        res = gpr.fit_residuals(self.coords, self.Y)
        self.assertLess(float(res.var()), float(self.Y.var()) * 0.8)

    def test_large_n_fixed_bounds(self):
        """Fixed-bounds large-n path should also reduce signal."""
        gpr = SklearnKernelGPR(
            constant_value=1.0,
            constant_value_bounds="fixed",
            length_scale_bounds="fixed",
            n_inducing=50,
        )
        res = gpr.fit_residuals(self.coords, self.Y)
        self.assertEqual(res.shape, self.Y.shape)

    def test_precompute_shared_kernel_warns_for_large_n(self):
        """precompute_shared_kernel should warn when n > n_inducing."""
        gpr = SklearnKernelGPR(
            constant_value_bounds="fixed",
            length_scale_bounds="fixed",
            n_inducing=50,
        )
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            gpr.precompute_shared_kernel(self.coords)  # n=300 > n_inducing=50
        large_n_warns = [x for x in w if "n_inducing" in str(x.message)]
        self.assertTrue(len(large_n_warns) > 0, "Expected warning about n_inducing")


@unittest.skipUnless(_GPYTORCH_AVAILABLE, "gpytorch not installed")
class TestGPyTorchKernelGPRSparse(unittest.TestCase):
    """Tests for the SGPR (InducingPointKernel) path in GPyTorchKernelGPR."""

    def setUp(self):
        torch.manual_seed(7)
        np.random.seed(7)
        self.coords = _make_coords(200)
        signal = torch.sin(self.coords[:, 0] * 2.0)
        self.Y = signal.unsqueeze(1) + 0.1 * torch.randn(200, 1)

    def test_sgpr_output_shape(self):
        gpr = GPyTorchKernelGPR(
            constant_value_bounds=(1e-3, 1e3),
            length_scale_bounds="fixed",
            n_inducing=50,
            n_iter=30,
        )
        res = gpr.fit_residuals(self.coords, self.Y)
        self.assertEqual(res.shape, self.Y.shape)

    def test_sgpr_reduces_signal(self):
        gpr = GPyTorchKernelGPR(
            constant_value_bounds=(1e-3, 1e3),
            length_scale_bounds="fixed",
            n_inducing=50,
            n_iter=50,
        )
        res = gpr.fit_residuals(self.coords, self.Y)
        self.assertLess(float(res.var()), float(self.Y.var()) * 0.8)

    def test_sgpr_nan_preserved(self):
        Y_nan = self.Y.clone()
        Y_nan[10] = float("nan")
        gpr = GPyTorchKernelGPR(
            constant_value=1.0,
            constant_value_bounds=(1e-2, 1e2),
            length_scale=1.0,
            length_scale_bounds="fixed",
            n_inducing=50,
            n_iter=20,
        )
        res = gpr.fit_residuals(self.coords, Y_nan)
        self.assertTrue(torch.isnan(res[10]).all())
        self.assertFalse(torch.isnan(res[0]).any())


if __name__ == "__main__":
    unittest.main()



"""Multinomial GLM/GLMM model implementations for SPLISOSM."""

import warnings
from abc import ABC, abstractmethod
from timeit import default_timer as timer
from typing import Optional, Literal

import torch
import torch.nn as nn
from torch.autograd.functional import jacobian
from torch.distributions import Gamma, InverseGamma

from splisosm.likelihood import (
    log_prob_fastmult_batched,
    log_prob_fastmvn_batched,
)
from splisosm.logger import PatienceLogger

__all__ = ["MultinomGLM", "MultinomGLMM"]


class BaseModel(ABC):
    """API for the GLM and GLMM model."""

    @abstractmethod
    def setup_data(
        self,
        counts: torch.Tensor,
        corr_sp: Optional[torch.Tensor],
        design_mtx: Optional[torch.Tensor] = None,
    ) -> None:
        """Set up the data for the model.

        Parameters
        ----------
        counts : torch.Tensor
            Shape (n_genes, n_spots, n_isoforms), genes with the same number of isoforms.
        corr_sp : torch.Tensor
            Shape (n_spots, n_spots).
        design_mtx : torch.Tensor, optional
            Shape (n_spots, n_factors).
        """
        pass

    def forward(self) -> torch.Tensor:
        """Calculate the log-likelihood or log-marginal-likelihood of the model."""
        pass

    @abstractmethod
    def fit(self) -> None:
        """Fit the model using all data."""
        pass

    @abstractmethod
    def get_isoform_ratio(self) -> torch.Tensor:
        """Extract the fitted isoform ratio across space."""
        pass

    @abstractmethod
    def clone(self) -> "BaseModel":
        """Clone the model with data and model parameters."""
        pass


def _melt_tensor_along_first_dim(tensor_in):
    """Melt a 4D tensor into 3D and reorder entries by spots.

    tensor_in[:, i, j, k] -> matrix_out[:, i + j * n, i + k * n] where n = tensor_in.shape[1]

    Parameters
    ----------
    tensor_in : torch.Tensor
        Shape (n_genes, n_spots, n_isos - 1, n_isos - 1).

    Returns
    -------
    matrix_out : torch.Tensor
        Shape (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1)).
    """
    b, n, m = tensor_in.shape[:3]
    assert tensor_in.shape == (b, n, m, m)

    # example: at spot i, isoform j and k are connected via tensor[i, j, k]
    # tensor[i, j, k] -> out[i + j * n_spots, i + k * n_spots]
    matrix_out = torch.zeros(b, n * m, n * m, device=tensor_in.device)
    i, j, k = torch.meshgrid(
        torch.arange(n), torch.arange(m), torch.arange(m), indexing="ij"
    )
    row_indices = i + j * n
    col_indices = i + k * n
    matrix_out[:, row_indices.view(-1), col_indices.view(-1)] += tensor_in.flatten(1)

    return matrix_out


@torch.no_grad
def update_at_idx(
    params: torch.Tensor, new_params: torch.Tensor, idx: torch.Tensor
) -> torch.Tensor:
    idx = idx.view(-1, *([1] * (params.ndim - 1))).float()
    params = params * idx + new_params * (1 - idx)
    return params


class MultinomGLM(BaseModel, nn.Module):
    """The Multinomial Generalized Linear Model for spatial isoform expression.

    Compared to MultinomGLMM, this model does not have a random effect term::

        Y ~ Multinomial(alpha, Y.sum(1))
        eta = multinomial-logit(alpha) = X @ beta + bias_eta

    Given isoform counts of a gene ``Y`` (n_spots, n_isos) and design matrix ``X`` (n_spots, n_factors),
    MultinomGLM.fit will find the MAP estimates of the following learnable parameters:

    - ``beta``: (n_factors, n_isos - 1) covariate coefficients of the fixed effect term.
    - ``bias_eta``: (n_isos - 1) intercepts of the fixed effect term.

    Inference is performed by maximizing the log likelihood using `fitting_method` (default: ``'iwls'``)

    Example
    -------
    >>> from splisosm.model import MultinomGLM
    >>> import torch
    >>> # Generate synthetic data
    >>> counts = torch.randint(0, 10, (5, 100, 3))  # 5 genes, 100 spots, each 3 isoforms
    >>> # Fit the GLM model
    >>> model = MultinomGLM(fitting_method='iwls')
    >>> model.setup_data(counts, design_mtx=None)
    >>> model.fit()
    >>> print(model)
    >>> # Extract the fitted isoform ratios
    >>> isoform_ratios = model.get_isoform_ratio()  # shape (5, 100, 3)
    >>> # Fitted parameters
    >>> print(model.beta.shape)  # shape (5, 0, 2)
    >>> print(model.bias_eta.shape)  # shape (5, 2)
    """

    n_spots: int
    """Number of samples/spots"""

    n_genes: int
    """Number of genes in the batch."""

    n_isos: int
    """Number of isoforms per gene in the batch."""

    n_factors: int | None
    """Number of covariates in the design matrix"""

    fitting_method: Literal["iwls", "newton", "gd"]
    """Method for fitting the model."""

    fitting_configs: dict
    """Dictionary of fitting configurations."""

    fitting_time: float
    """Time taken for fitting the model."""

    def __init__(
        self,
        fitting_method: Literal["iwls", "newton", "gd"] = "iwls",
        fitting_configs: dict = {},
    ):
        """
        Parameters
        ----------
        fitting_method
            Method for fitting the model.
            ``'iwls'``: Iteratively reweighted least squares.
            ``'newton'``: Newton's method.
            ``'gd'``: Gradient descent.
        fitting_configs
            Dictionary of fitting configurations. Keys include

            - ``'lr'``: float, Learning rate for gradient descent or Newton's method.
            - ``'optim'``: str, Optimizer type, one of ``'adam'``, ``'sgd'``, or ``'lbfgs'``.
            - ``'tol'``: float, Tolerance for convergence.
            - ``'max_epochs'``: int, Maximum number of epochs for fitting.
            - ``'patience'``: int, Number of epochs to wait for improvement before stopping.
        """
        super().__init__()
        # will be set up later by calling self.setup_data()
        self.n_spots = None  # number of spots
        self.n_isos = None  # number of isoforms
        self.n_factors = None  # number of covariates

        # specify the fitting method
        assert fitting_method in ["gd", "newton", "iwls"]
        self.fitting_method = fitting_method

        # specify the fitting configurations
        self.fitting_configs = {  # default configurations
            "lr": 1e-2,
            "optim": "adam",
            "tol": 1e-5,
            "max_epochs": 1000,
            "patience": 5 if fitting_method == "gd" else 2,
        }
        self.fitting_configs.update(fitting_configs)

        # for now, restricting the optimization method to Adam, SGD and lbfgs
        assert self.fitting_configs["optim"] in ["adam", "sgd", "lbfgs"]

        # specify the fitting outcomes
        self.fitting_time = 0

    def __str__(self):
        return (
            "A Multinomial Generalized Linear Model (GLM)\n"
            + f"- Number of genes in the batch: {self.n_genes}\n"
            + f"- Number of spots: {self.n_spots}\n"
            + f"- Number of isoforms per gene: {self.n_isos}\n"
            + f"- Number of covariates: {self.n_factors}\n"
            + f"- Fitting method: {self.fitting_method}"
        )

    def setup_data(
        self,
        counts: torch.Tensor,
        design_mtx: Optional[torch.Tensor] = None,
        device: Literal["cpu", "cuda"] = "cpu",
    ) -> None:
        """Set up the data for the model.

        Parameters
        ----------
        counts
            Shape (n_genes, n_spots, n_isos) or (n_spots, n_isos).
            For batched calculations, all genes in the batch must have the same number of isoforms.
        design_mtx
            Shape (n_spots, n_factors). Design matrix of spatial covariates.
            If None, an intercept-only design matrix will be used.
        device
            'cpu' or 'cuda'. 'mps' currently not supported (torch.lgamma not supported on mps).
        """
        # need to switch to a different count model (e.g., Poisson) when only one isoform is provided
        if counts.shape[-1] == 1:
            raise NotImplementedError(
                "Only one isoform provided. Please use a different count model."
            )

        assert device in ["cpu", "cuda"]
        self.device = torch.device(device)

        if counts.ndim == 2:
            counts = counts.unsqueeze(0)  # (1, n_spots, n_isos)
            print(
                "Batched calculation has been implemented. Provide a batch of counts to speed up calculation."
            )

        # convert sparse tensors to dense tensors
        if counts.is_sparse:
            counts = counts.to_dense()

        # set model dimensions based on the input shape
        self.n_genes, self.n_spots, self.n_isos = counts.shape

        if design_mtx is None:
            # initialize an empty design matrix of shape (1, n_spots, 0)
            design_mtx = torch.ones(1, self.n_spots, 0)
        elif design_mtx.ndim == 2:
            design_mtx = design_mtx.unsqueeze(0)  # (1, n_spots, n_factors)
        elif design_mtx.ndim == 3:
            # all genes to test should share the same design matrix
            assert (
                design_mtx.shape[0] == 1
            ), "Batched design matrix is currently not supported."
        else:
            raise ValueError(
                f"design_mtx must be a 2D or 3D tensor. Got shape: {design_mtx.shape}"
            )
        self.n_factors = design_mtx.shape[-1]
        assert design_mtx.shape == (
            1,
            self.n_spots,
            self.n_factors,
        ), f"Invalide design matrix shape: {design_mtx.shape}"

        # (n_genes, n_spots, n_isos), int, the observed counts
        self.register_buffer("counts", counts)
        # (1, n_spots, n_factors), the input design matrix of n_factors covariates
        self.register_buffer("X_spot", design_mtx)
        self.register_buffer("convergence", torch.zeros(self.n_genes, dtype=bool))

        # set up learnable parameters according to the model architecture
        self._configure_learnable_variables()

        # send to device
        self.to(self.device)

    def _configure_learnable_variables(self, val=None):
        """Set up learnable parameters according to the model architecture."""
        # the fixed effect terms
        # covariate coefficients
        beta = torch.zeros(self.n_genes, self.n_factors, self.n_isos - 1)
        # intercepts
        bias_eta = torch.zeros(self.n_genes, self.n_isos - 1)

        # optimize using gradient descent / newton's method
        self.register_parameter("beta", nn.Parameter(beta))
        self.register_parameter("bias_eta", nn.Parameter(bias_eta))

    def _configure_optimizer(self, verbose=False):
        """Configure the optimizer and learning rate scheduler."""
        # initialize optimizer
        if self.fitting_configs["optim"] == "adam":
            self.optimizer = torch.optim.Adam(
                self.parameters(), lr=self.fitting_configs["lr"]
            )
            self._closure = None
        elif self.fitting_configs["optim"] == "sgd":
            self.optimizer = torch.optim.SGD(
                self.parameters(), lr=self.fitting_configs["lr"]
            )
            self._closure = None
        else:
            # with torch.no_grad():
            #     def _fake_parameters():
            #         for p in self.parameters():
            #             yield torch.tensor(
            #                 p.flatten(end_dim=1), requires_grad=True, device=p.device
            #             )
            #     self._fake_parameters = _fake_parameters
            if hasattr(self, "_fake_params"):
                del self._fake_params
            self.register_module(
                "_fake_params",
                nn.ParameterList(
                    [
                        nn.ParameterList(
                            [nn.Parameter(p[i].flatten()) for p in self.parameters()]
                        )
                        for i in range(self.n_genes)
                    ]
                ),
            )
            # from here on, no new parameters should be added
            self.optimizers = [
                torch.optim.LBFGS(
                    self._fake_params[i],
                    lr=self.fitting_configs["lr"],
                    max_iter=10,
                    tolerance_change=self.fitting_configs["tol"],
                    line_search_fn="strong_wolfe",
                )
                for i in range(self.n_genes)
            ]

            def closure():
                i = self._temp_mask
                self.optimizers[i].zero_grad()
                with torch.no_grad():
                    for p, fp in zip(self.parameters(), self._fake_params[i]):
                        # print(p.shape, fp.shape)
                        if p.grad is not None:
                            p.grad[i].zero_()
                        else:
                            p.grad = torch.zeros_like(p)
                        # copy value from fake to real
                        p[i].data.copy_(fp.data.reshape_as(p[i]))
                neg_log_prob = -self()[i].sum()
                neg_log_prob.backward()
                # print(self.beta.grad)
                with torch.no_grad():  # copy gradient from real to fake
                    for p, fp in zip(self.parameters(), self._fake_params[i]):
                        if fp.grad is None:
                            fp.grad = torch.zeros_like(fp)
                        fp.grad.copy_(p.grad[i].reshape_as(fp))
                return neg_log_prob

            self._closure = closure

        # learning rate scheduler
        # self.scheduler = ReduceLROnPlateau(
        #     self.optimizer,
        #     patience=int(self.fitting_configs["patience"] / 4) + 1,
        #     factor=0.1,
        #     verbose=verbose,
        # )

    def _eta(self):
        """Output the eta based on the linear model."""
        # (1, n_spots, n_factors) @ (n_genes, n_factors, n_isos - 1) + (n_genes, 1, n_isos - 1)
        # -> (n_genes, n_spots, n_isos - 1)
        return self.X_spot.matmul(self.beta) + self.bias_eta.unsqueeze(-2)

    def _alpha(self):
        """Convert eta (n_isos - 1) to alpha (n_isos)."""
        # alpha is the expected proportion of isoforms
        # alpha.shape = (n_genes, n_spots, n_isos)
        # the last isoform will have constant zero eta across space and
        # its proportion is given by 1/(1 + sum(exp(eta)))
        eta = self._eta()  # (n_genes, n_spots, n_isos - 1)
        alpha = torch.cat(
            [eta, torch.zeros(self.n_genes, self.n_spots, 1, device=self.device)],
            dim=-1,
        )  # (n_genes, n_spots, n_isos)
        alpha = torch.softmax(alpha, dim=-1)  # (n_genes, n_spots, n_isos)
        return alpha

    def get_isoform_ratio(self) -> torch.Tensor:
        """Extract the fitted isoform ratio across space.

        Returns
        -------
        ratio : torch.Tensor
            Shape (n_genes, n_spots, n_isos), the fitted isoform ratio across space.
        """
        return self._alpha().detach()  # (n_genes, n_spots, n_isos)

    def forward(self) -> torch.Tensor:
        """Calculate log probability given data.

        Returns
        -------
        log_prob : torch.Tensor
            Shape (n_genes,), the log probability for each gene.
        """
        return log_prob_fastmult_batched(
            self._alpha().transpose(1, 2), self.counts.transpose(1, 2)
        )

    # Functions to calculate Hessian.

    def _get_log_lik_gradient_beta_bias(self):
        """Get the gradient of the log joint probability wrt beta and bias."""
        # calculate the gradient wrt eta
        # (n_genes, n_spots, n_isos)
        d_l_d_eta = self.counts - self._alpha() * self.counts.sum(axis=-1, keepdim=True)

        # calculate the gradient wrt beta and bias
        X_expand = torch.cat(
            [
                self.X_spot,
                torch.ones(1, self.n_spots, 1, device=self.device),
            ],
            dim=-1,
        )  # (1, n_spots, n_factors + 1)
        # score of shape (n_genes, n_factors + 1, n_isos - 1)
        score = X_expand.transpose(1, 2).matmul(d_l_d_eta.detach()[..., :-1])

        return score

    def _get_log_lik_hessian_eta(self):
        """Get the Hessian matrix of the log joint probability wrt eta."""
        n_isos = self.n_isos
        props = self._alpha()  # (n_genes, n_spots, n_isos)

        # multinom_hessian[s] = - sum(counts[s,:]) * {diag(p) - p@p.T}
        # (n_genes, n_spots, n_isos - 1, n_isos - 1)
        multinom_hessian = -self.counts.sum(-1).view(self.n_genes, -1, 1, 1) * (
            props[..., :-1].unsqueeze(-1).expand(-1, -1, -1, n_isos - 1)
            * torch.eye(n_isos - 1, device=self.device)
            - torch.einsum("bsi,bsj->bsij", (props[..., :-1], props[..., :-1]))
        )

        # reshape the hessian into (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        # (1) zeros for spots i != j
        # (2) at spot i, isoform j and k are connected via multinom_hessian[i, j, k]
        return _melt_tensor_along_first_dim(multinom_hessian)

    def _get_log_lik_hessian_beta_bias(self):
        """Get the Hessian matrix of the log joint probability wrt the fixed effects."""
        # combine beta and the intercept bias_eta
        X_expand = torch.cat(
            [
                self.X_spot,
                torch.ones(1, self.n_spots, 1, device=self.device),
            ],
            dim=-1,
        )  # (1, n_spots, n_factors + 1)
        # (n_genes, n_spots * (n_isos - 1), (n_factors + 1) * (n_isos - 1))
        X_expand_full = torch.stack(
            [torch.block_diag(*[x] * (self.n_isos - 1)) for x in X_expand], dim=0
        )

        # calculate the hessian of the log likelihood wrt [beta, bias]
        # W is the hessian of the log multinomial likelihood wrt eta := X@beta + nu
        # W[s] = - sum(counts[s,:]) * {diag(p) - p@p.T}
        # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        W = self._get_log_lik_hessian_eta()
        # hessian_beta = X.T @ W @ X, ((n_factors + 1)(n_isos - 1), (n_factors + 1)(n_isos - 1))
        # (n_genes, (n_factors + 1) * (n_isos - 1), (n_factors + 1) * (n_isos - 1))
        hessian_beta_expand = (
            X_expand_full.transpose(1, 2).matmul(W).matmul(X_expand_full)
        )

        return hessian_beta_expand

    """Functions for model fitting.
    """

    def _update_gradient_descent(self):
        """Update the model parameters using gradient descent."""
        if self.fitting_configs["optim"] == "lbfgs":
            [optimizer.zero_grad() for optimizer in self.optimizers]
            with torch.no_grad():
                for p in self.parameters():
                    if p.grad is not None:
                        p.grad.zero_()
        else:
            self.optimizer.zero_grad()

        # minimize the negative log-likelihood or the negative log-marginal-likelihood
        neg_log_prob = -self()[~self.convergence].sum()
        neg_log_prob.backward()

        # gradient-based updates
        if self.fitting_configs["optim"] == "lbfgs":
            # update all parameters using L-BFGS
            [optimizer.zero_grad() for optimizer in self.optimizers]
            with torch.no_grad():
                for p in self.parameters():
                    if p.grad is not None:
                        p.grad.zero_()
            for idx, optimizer in enumerate(self.optimizers):
                self._temp_mask = idx
                optimizer.step(self._closure)
            with torch.no_grad():
                for idx, fps in enumerate(self._fake_params):
                    for p, fp in zip(self.parameters(), fps):
                        p[idx].data.copy_(fp.data.reshape_as(p[idx]))
            # print(self.beta[0], self.bias_eta[0])
        elif self.fitting_configs["optim"] in ["adam", "sgd"]:
            # update the remaining parameters with non-zero gradients using gradient descent
            self.optimizer.step()
        else:
            raise NotImplementedError(
                f"Optimization method {self.fitting_configs['optim']} is not supported."
            )

    def _update_newton(self, step=0.9):
        """Update the model parameters using Newton's method."""
        n_genes, n_isos, n_factors = self.n_genes, self.n_isos, self.n_factors

        # combine beta and bias_eta
        # (n_genes, n_factors, n_isos - 1), (n_genes, 1, n_isos - 1) -> (n_genes, n_factors + 1, n_isos - 1)
        beta_expand = torch.cat([self.beta, self.bias_eta.unsqueeze(1)], dim=1)

        # calculate gradient and hessian
        # gradient_beta_expand = torch.cat([self.beta.grad, self.bias_eta.grad.reshape(1,-1)], dim=0)
        # (n_genes, n_factors + 1, n_isos - 1)
        gradient_beta_expand = self._get_log_lik_gradient_beta_bias()
        # (n_genes, (n_factors + 1) * (n_isos - 1), (n_factors + 1) * (n_isos - 1))
        hessian_beta_expand = self._get_log_lik_hessian_beta_bias()
        # for numerical stability
        # (n_genes, (n_factors + 1) * (n_isos - 1), (n_factors + 1) * (n_isos - 1))
        hessian_beta_expand += 1e-5 * torch.eye(
            (n_factors + 1) * (n_isos - 1), device=self.device
        )

        # find the new beta and bias_eta using the Newton's method
        # (n_genes, (n_factors + 1) * (n_isos - 1))
        right = torch.einsum(
            "bij,bj->bi", hessian_beta_expand, beta_expand.transpose(1, 2).flatten(1)
        ) - step * gradient_beta_expand.transpose(1, 2).flatten(1)
        beta_expand_new = (
            torch.linalg.solve(hessian_beta_expand, right)
            .reshape(n_genes, n_isos - 1, n_factors + 1)
            .transpose(1, 2)
        )

        # extract beta and bias_eta
        beta_new = beta_expand_new[:, :-1, :]  # (n_genes, n_factors, n_isos - 1)
        bias_eta_new = beta_expand_new[:, -1, :]  # (n_genes, n_isos - 1)

        # update the parameters and clear the gradients
        self.beta.data.copy_(update_at_idx(self.beta, beta_new, self.convergence))
        self.bias_eta.data.copy_(
            update_at_idx(self.bias_eta, bias_eta_new, self.convergence)
        )

    def _update_iwls(self):
        """Update the model parameters using the iteratively reweighted least squares (IWLS)."""
        n_genes, n_spots, n_isos, n_factors = (
            self.n_genes,
            self.n_spots,
            self.n_isos,
            self.n_factors,
        )
        props = self._alpha()  # (n_genes, n_spots, n_isos)

        # define the working variable (pseudo data)
        # working_y = eta + [d_eta/d_mu][counts - mu]
        # [d_eta/d_mu] = [d_mu/d_eta]^-1 = W^-1
        # W = N_i * (diag(alpha[i,:-1]) - alpha[i,:-1].T @ alpha[i,:-1]) at given location i

        # W = sum(counts[s,:]) * {diag(p) - p@p.T}
        # (n_genes, n_spots, n_isos - 1, n_isos - 1)
        W = self.counts.sum(-1).view(n_genes, n_spots, 1, 1) * (
            props[..., :-1].unsqueeze(-1).expand(-1, -1, -1, n_isos - 1)
            * torch.eye(n_isos - 1, device=self.device)
            - torch.einsum("bsi,bsj->bsij", (props[..., :-1], props[..., :-1]))
        )
        # for stability
        W_inv = torch.linalg.inv(W + 1e-5 * torch.eye(n_isos - 1, device=self.device))
        residuals = self.counts - props * self.counts.sum(-1, keepdim=True)
        working_y = residuals[..., :-1].unsqueeze(
            -1
        )  # (n_genes, n_spots, n_isos - 1, 1)
        working_y = (W_inv.matmul(working_y)).squeeze(
            -1
        )  # (n_genes, n_spots, n_isos - 1)
        working_y += self._eta()  # (n_genes, n_spots, n_isos - 1)

        # calculate the IWLS update for beta and bias_eta
        # beta_new = (X.T @ W @ X)^-1 @ X.T @ W @ y
        # reshape W
        # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        W = _melt_tensor_along_first_dim(W)
        # reshape X to include the intercept term
        # (1, n_spots, n_factors + 1)
        X_expand = torch.cat(
            [self.X_spot, torch.ones(1, n_spots, 1, device=self.device)], dim=-1
        )
        # (1, n_spots * (n_isos - 1), (n_factors + 1) * (n_isos - 1))
        X_expand_full = torch.stack(
            [torch.block_diag(*[x] * (n_isos - 1)) for x in X_expand], dim=0
        )
        # calculate the analytical solution
        # (n_genes, (n_factors + 1) * (n_isos - 1), n_spots * (n_isos - 1))
        Xt_W = X_expand_full.transpose(1, 2).matmul(W)
        # (n_genes, (n_factors + 1) * (n_isos - 1), (n_factors + 1) * (n_isos - 1))
        Xt_W_X = Xt_W.matmul(X_expand_full)
        # add a small value to the diagonal for stability
        Xt_W_X += 1e-5 * torch.eye((n_factors + 1) * (n_isos - 1))

        # (n_genes, (n_factors + 1) x (n_isos - 1))
        Xt_W_y = torch.einsum("bij,bj->bi", Xt_W, working_y.transpose(1, 2).flatten(1))
        res = torch.linalg.solve(Xt_W_X, Xt_W_y).reshape(
            n_genes, n_isos - 1, n_factors + 1
        )

        # extract beta and bias_eta
        beta_new = res[..., :-1].transpose(1, 2)  # (n_genes, n_factors, n_isos - 1)
        bias_eta_new = res[..., -1]  # (n_genes, n_isos - 1)

        # update the parameters and clear the gradients
        # with torch.no_grad():
        #     self.beta[~self.convergence].copy_(beta_new[~self.convergence])
        #     self.bias_eta[~self.convergence].copy_(bias_eta_new[~self.convergence])
        self.beta.data.copy_(update_at_idx(self.beta, beta_new, self.convergence))
        self.bias_eta.data.copy_(
            update_at_idx(self.bias_eta, bias_eta_new, self.convergence)
        )

    def fit(
        self,
        diagnose: bool = False,
        verbose: bool = False,
        quiet: bool = False,
        random_seed: Optional[int] = None,
    ) -> dict:
        """Fit the model using all data

        Parameters
        ----------
        diagnose
            Whether to store parameter changes during training (passed to PatienceLogger).
        verbose
            Whether to print verbose information during fitting.
        quiet
            Whether to suppress output during fitting.
        random_seed
            Random seed for reproducibility.

        Returns
        -------
        params_iter : dict or None
            If `diagnose` is True, returns a dictionary of parameter changes during training. Otherwise returns None.
        """
        if random_seed is not None:  # set random seed for reproducibility
            torch.manual_seed(random_seed)

        if self.fitting_method == "gd":  # use gradient descent
            # configure the optimizer and start training
            self._configure_optimizer(verbose=verbose)
            self.train()

        max_epochs = self.fitting_configs["max_epochs"]
        patience = self.fitting_configs["patience"]
        tol = self.fitting_configs["tol"]
        max_epochs = 10000 if max_epochs == -1 else max_epochs
        patience = patience if patience > 0 else 1

        # set iteration limits
        batch_size = self.n_genes
        t_start = timer()
        logger = PatienceLogger(batch_size, patience, min_delta=tol, diagnose=diagnose)

        while logger.epoch < max_epochs and not logger.convergence.all():
            # update the model parameters
            if self.fitting_method == "gd":
                self._update_gradient_descent()
            elif self.fitting_method == "newton":
                self._update_newton()
            elif self.fitting_method == "iwls":
                self._update_iwls()

            # check convergence
            with torch.no_grad():
                # calculate the negative log-likelihood
                neg_log_prob = -self().detach().cpu()

                # d_loss = prev_loss - neg_log_prob.detach().item()

                # # update the epoch and patience
                # epoch += 1
                # if d_loss < 0:  # if loss increases
                #     patience -= 1

                # # update the loss
                # prev_loss = neg_log_prob.detach().item()

            # if self.fitting_method == "gd":
            # update learning rate
            # self.scheduler.step(neg_log_prob[~self.convergence].mean())

            logger.log(
                neg_log_prob,
                {
                    k: v.detach().cpu()
                    for k, v in self.named_parameters()
                    if "_fake" not in k
                },
            )
            self.convergence.copy_(logger.convergence)

            if (verbose and not quiet) and logger.epoch % 10 == 0:
                print(
                    f"Epoch {logger.epoch}. Loss (neg_log_prob): {logger.best_loss.mean():.4f}. "
                )

        # check model convergence
        num_not_converge = (~logger.convergence).sum()
        if num_not_converge:
            warnings.warn(
                f"{num_not_converge} Genes did not converge after epoch {max_epochs}. "
                "Try larger max_epochs."
            )

        # save runtime
        t_end = timer()
        self.fitting_time = t_end - t_start

        if not quiet:  # print final message
            print(
                f"Time {self.fitting_time:.2f}s. Total epoch {logger.epoch}. Final loss "
                f"(neg_log_prob): {neg_log_prob.mean():.3f}."
            )

        # collect parameters corresponding to the best epoch for each sample in batch
        if max_epochs > 0:
            for k, v in self.named_parameters():
                if "_fake" not in k:
                    v.data.copy_(logger.best_params[k])

        self.logger = logger

        return logger.params_iter

    def clone(self) -> "MultinomGLM":
        """Clone a model with the same set of parameters."""
        new_model = type(self)(
            fitting_method=self.fitting_method, fitting_configs=self.fitting_configs
        )
        new_model.setup_data(counts=self.counts, design_mtx=self.X_spot)
        new_model.load_state_dict(self.state_dict())

        return new_model

    def update_params_from_dict(self, params: dict) -> None:
        """Update a subset of model parameters with a dictionary of parameters.

        Parameters
        ----------
        params
            A dictionary of parameters to be updated. The keys must be
            existing parameter names in the model.
        """
        new_params = self.state_dict()
        new_params.update(params)
        self.load_state_dict(new_params)


class MultinomGLMM(MultinomGLM, BaseModel, nn.Module):
    """The Multinomial Generalized Linear Mixed Model for spatial isoform expression.

    The model is defined as follows::

        Y ~ Multinomial(alpha, Y.sum(1))
        eta = multinomial-logit(alpha) = X @ beta + bias_eta + nu
        nu ~ MVN(0, sigma^2 * (theta * V_sp + (1-theta) * I) =
             MVN(0, sigma_sp^2 * V_sp + sigma_nsp^2 * I)

    Given isoform counts of a gene ``Y`` (n_spots, n_isos), design matrix ``X`` (n_spots, n_factors),
    and spatial covariance matrix ``V_sp`` (n_spots, n_spots), the model estimates the isoform usage
    ratio ``alpha`` (n_spots, n_isos) across space.
    Specifically, `MultinomGLMM.fit` will find the MAP estimates of the following learnable parameters:

    - ``beta``: (n_factors, n_isos - 1) covariate coefficients of the fixed effect term.
    - ``bias_eta``: (n_isos - 1) intercepts of the fixed effect term.
    - ``nu``: (n_spots, n_isos - 1) the random effect term.
    - variance components: each of length n_isos - 1 (or 1 if `share_variance` is True).
      If `var_parameterization_sigma_theta` is True, they are (``sigma``, ``theta_logit``),
      representing total variance and logit of spatial variance proportion ``theta``.
      Otherwise they are (``sigma_sp``, ``sigma_nsp``), representing spatial and non-spatial
      variance components.

    Inference algorithms can be categorized into two types based on the optimization objective:

    - Joint: Maximize the joint likelihood (with the random effect ``nu``).
      This is equivalent to the first-order Laplace approximation of the marginal likelihood.
    - Marginal: Maximize the marginal likelihood (with the random effect ``nu`` integrated out).
      The integral is approximated by a second-order Laplace approximation.

    Methods implemented:

    - ``'joint_gd'``: Maximize the joint likelihood using gradient descent.
    - ``'joint_newton'``: Maximize the joint likelihood using Newton's method.
    - ``'marginal_gd'``: Maximize the marginal likelihood using gradient descent.
    - ``'marginal_newton'``: Maximize the marginal likelihood using Newton's method.
      In this method, ``nu`` is first updated using Newton's method every
      ``'update_nu_every_k'`` iterations, and ``beta``, ``bias_eta``, and variance components
      are updated using gradient descent.

    Notes
    -----
    It is also possible to implement held-out likelihood for model selection.

    Example
    -------
    >>> from splisosm.model import MultinomGLMM
    >>> from splisosm.utils import get_cov_sp
    >>> import torch
    >>> # Generate synthetic data
    >>> counts = torch.randint(0, 10, (5, 100, 3))  # 5 genes, 100 spots, each 3 isoforms
    >>> coords = torch.rand(100, 2)  # 100 spots with 2D coordinates
    >>> K_sp = get_cov_sp(coords, k=4, rho=0.9) # spatial covariance matrix of shape (100, 100)
    >>> # Fit the GLMM model
    >>> model = MultinomGLMM(fitting_method='joint_gd')
    >>> model.setup_data(counts, corr_sp=K_sp, design_mtx=None)
    >>> model.fit()
    >>> print(model)
    >>> # Extract the fitted isoform ratios
    >>> isoform_ratios = model.get_isoform_ratio()  # shape (5, 100, 3)
    >>> # Fitted parameters
    >>> print(model.beta.shape)  # shape (5, 0, 2)
    >>> print(model.bias_eta.shape)  # shape (5, 2)
    >>> print(model.nu.shape)  # shape (5, 100, 2)
    >>> print(model.sigma.shape)  # shape (5, 1)
    >>> print(model.theta_logit.shape)  # shape (5, 1)
    """

    share_variance: bool
    """Whether to use the same variance across isoforms."""

    var_parameterization_sigma_theta: bool
    """Whether variance components are parameterized as (``sigma``, ``theta_logit``) or (``sigma_sp``, ``sigma_nsp``)."""

    var_fix_sigma: bool
    """Whether to fix the total variance (``sigma``) or not."""

    var_prior_model: Literal["none", "gamma", "inv_gamma"]
    """The prior model on the total variance ``sigma``."""

    var_prior_model_params: dict
    """The parameters for the prior model on the total variance ``sigma``."""

    init_ratio: Literal["observed", "uniform"]
    """The initialization method for the logit isoform usage ratio ``gamma``."""

    fitting_method: Literal[
        "joint_gd", "joint_newton", "marginal_gd", "marginal_newton"
    ]
    """The fitting method to use."""

    fitting_configs: dict
    """A dictionary of fitting configurations."""

    fitting_time: float
    """The time taken to fit the model."""

    def __init__(
        self,
        share_variance: bool = True,
        var_parameterization_sigma_theta: bool = True,
        var_fix_sigma: bool = True,
        var_prior_model: Literal["none", "gamma", "inv_gamma"] = "none",
        var_prior_model_params: dict = {},
        init_ratio: Literal["observed", "uniform"] = "observed",
        fitting_method: Literal[
            "joint_gd", "joint_newton", "marginal_gd", "marginal_newton"
        ] = "joint_gd",
        fitting_configs: dict = {},
    ):
        """
        Parameters
        ----------
        share_variance
            Whether to use the same variance across isoforms. If True, the variance components
            will be of length 1. If False, the variance components will be of length n_isos - 1.
        var_parameterization_sigma_theta
            Whether to parameterize the variance components as (``sigma``, ``theta_logit``) or (``sigma_sp``, ``sigma_nsp``).
            If True, the variance components will be (``sigma``, ``theta_logit``), where ``sigma`` is the total variance and
            ``theta_logit`` is the logit of the spatial variance proportion.
            If False, the variance components will be (``sigma_sp``, ``sigma_nsp``), where ``sigma_sp`` is the spatial
            variance and ``sigma_nsp`` is the non-spatial variance.
        var_fix_sigma
            Whether to fix the total variance (``sigma``) or not. If True, the total variance will be fixed to the initial value,
            which is the average per-spot variance of isoform counts normalized by its mean expression.
            See `MultinomGLMM._initialize_params` for details.
        var_prior_model
            The prior model on the total variance ``sigma``. Default is ``'none'`` with no prior.
            Other options are ``'gamma'`` (Gamma prior) and ``'inv_gamma'`` (Inverse Gamma prior).
        var_prior_model_params
            The parameters for the prior model on the total variance ``sigma``.
            For ``'gamma'``, the default parameters are ``{'alpha': 2.0, 'beta': 0.3}``.
            For ``'inv_gamma'``, the default parameters are ``{'alpha': 3, 'beta': 0.5}``.
        init_ratio
            The initialization method for the logit isoform usage ratio. Options are ``'observed'`` (initialize using observed counts)
            and ``'uniform'`` (equal isoform usage across space).
        fitting_method
            The fitting method to use. Options are ``'joint_gd'`` (joint likelihood with gradient descent),
            ``'joint_newton'`` (joint likelihood with Newton's method),
            ``'marginal_gd'`` (marginal likelihood with gradient descent),
            and ``'marginal_newton'`` (marginal likelihood with Newton's method).
        fitting_configs
            A dictionary of fitting configurations with the following keys:

            - ``'lr'``: float, learning rate
            - ``'optim'``: str, optimization method (one of ``'adam'``, ``'sgd'``, or ``'lbfgs'``)
            - ``'tol'``: float, tolerance for convergence
            - ``'max_epochs'``: int, maximum number of epochs
            - ``'patience'``: int, number of epochs to wait for improvement before stopping
            - ``'update_nu_every_k'``: int, number of iterations to update ``nu`` when using ``fitting_method='marginal_newton'``
        """
        super().__init__()

        # specify the parameterization of variance components
        # if True:
        # 	var(sigma, theta) = sigma^2 (theta * V_sp + (1-theta) * I)
        # else:
        # 	var(sigma_sp, sigma_nsp) = sigma_sp^2 * V_sp + sigma_nsp^2 * I
        self.var_parameterization_sigma_theta = var_parameterization_sigma_theta
        self.share_variance = (
            share_variance  # whether to share variance across isoforms
        )
        self.var_fix_sigma = var_fix_sigma  # whether to fix sigma

        # specify the prior model on sigma^2 (or sigma_sp^2 + sigma_nsp^2))
        assert var_prior_model in ["none", "gamma", "inv_gamma"]
        self.var_prior_model = var_prior_model  # prior on the variance size sigma
        if self.var_prior_model == "gamma":
            # Chung, Yeojin, et al. Psychometrika 78.4 (2013): 685-709.
            # this prior is applied on sigma (or sqrt(sigma_sp^2 + sigma_nsp^2))
            # Gamma(2, 0.3): prior mode of sigma ~= 3
            self.var_prior_model_params = {
                "alpha": 2.0,
                "beta": 0.3,
            }
            self.var_prior_model_params.update(var_prior_model_params)
            self.var_prior_model_dist = Gamma(
                self.var_prior_model_params["alpha"],
                self.var_prior_model_params["beta"],
            )
        elif self.var_prior_model == "inv_gamma":
            # conjugacy prior
            # this prior is applied on sigma^2 (or sigma_sp^2 + sigma_nsp^2)
            # InverseGamma(3, 0.5): prior mode of sigma^2 ~= 0.25
            self.var_prior_model_params = {
                "alpha": 3,
                "beta": 0.5,
            }
            self.var_prior_model_params.update(var_prior_model_params)
            self.var_prior_model_dist = InverseGamma(
                self.var_prior_model_params["alpha"],
                self.var_prior_model_params["beta"],
            )
        else:  # 	no/flat prior on sigma
            self.var_prior_model_params = {}
            self.var_prior_model_dist = None

        # specify the initialization method for the logit isoform usage ratio gamma
        assert init_ratio in ["observed", "uniform"]
        self.init_ratio = init_ratio

        # specify the fitting method
        assert fitting_method in [
            "joint_gd",
            "joint_newton",  # joint likelihood
            "marginal_gd",
            "marginal_newton",  # marginal likelihood
        ]
        self.fitting_method = fitting_method

        # specify the fitting configurations
        self.fitting_configs = {  # default configurations
            "lr": 1e-2,
            "optim": "adam",
            "tol": 1e-5,
            "max_epochs": 1000,
            "patience": 5,
        }
        self.fitting_configs.update(fitting_configs)
        if self.fitting_method == "joint_newton":
            # Newton's method is fast but can't very well handel saddle points
            # use small patience to avoid loss increase in the final iterations
            self.fitting_configs["patience"] = 2
            assert (
                self.var_fix_sigma is False
            ), "Newton's method requires sigma to be optimized."

        elif self.fitting_method == "marginal_gd":
            self.fitting_configs["lr"] = 1e-1

        elif self.fitting_method == "marginal_newton":
            # update nu using Newton's method every 'update_nu_every_k' iterations
            # update beta, bias_eta, and variance components using gradient descent
            self.fitting_configs["lr"] = 1e-1
            self.fitting_configs["patience"] = 10
            self.fitting_configs["update_nu_every_k"] = 3

        # override the default if user provides a different configuration
        self.fitting_configs.update(fitting_configs)

        # for now, restricting the optimization method to Adam, SGD and lbfgs
        assert self.fitting_configs["optim"] in ["adam", "sgd", "lbfgs"]

        # specify the fitting outcomes
        self.fitting_time = 0

    def __str__(self):
        return (
            "A Multinomial Generalized Linear Mixed Model (GLMM)\n"
            + f"- Number of genes in the batch: {self.n_genes}\n"
            + f"- Number of spots: {self.n_spots}\n"
            + f"- Number of isoforms per gene: {self.n_isos}\n"
            + f"- Number of covariates: {self.n_factors}\n"
            + "- Variance formulation:\n"
            + f"\t* Parameterized using sigma and theta: {self.var_parameterization_sigma_theta}\n"
            + f"\t* Learnable variance: {not self.var_fix_sigma}\n"
            + f"\t* Same variance across isoforms: {self.share_variance}\n"
            + f"\t* Prior on total variance: {self.var_prior_model}\n"
            + f"- Initialization method: {self.init_ratio}\n"
            + f"- Fitting method: {self.fitting_method}"
        )

    def setup_data(
        self,
        counts: torch.Tensor,
        corr_sp: Optional[torch.Tensor] = None,
        design_mtx: Optional[torch.Tensor] = None,
        device: Literal["cpu", "cuda"] = "cpu",
        corr_sp_eigvals: Optional[torch.Tensor] = None,
        corr_sp_eigvecs: Optional[torch.Tensor] = None,
    ) -> None:
        """Set up the data for the model.

        Parameters
        ----------
        counts
            Shape (n_genes, n_spots, n_isoforms) or (n_spots, n_isoforms).
            For batched calculations, all genes in the batch must have the same number of isoforms.
        corr_sp
            Shape (n_spots, n_spots), spatial covariance matrix.
            If None, the eigendecomposition of the spatial covariance matrix must be provided.
        design_mtx
            Shape (n_spots, n_factors). Design matrix of spatial covariates.
            If None, an intercept-only design matrix will be used.
        device
            'cpu' or 'cuda'. 'mps' currently not supported (torch.lgamma not supported on mps).
        corr_sp_eigvals
            Shape (n_spots,), eigenvalues of spatial covariance.
            If None, the spatial covariance matrix `corr_sp` must be provided.
        corr_sp_eigvecs
            Shape (n_spots, n_spots), eigenvectors of spatial covariance.
            If None, the spatial covariance matrix `corr_sp` must be provided.
        """
        # need to switch to a different count model (e.g., Poisson) when only one isoform is provided

        assert device in ["cpu", "cuda"]
        self.device = torch.device(device)

        if counts.ndim == 2:
            counts = counts.unsqueeze(0)  # (1, n_spots, n_isos)
            print(
                "Batched calculation has been implemented. Provide a batch of counts to speed up calculation."
            )
        else:
            if not counts.ndim == 3:
                raise ValueError(
                    f"counts must be a 2D or 3D tensor. Got shape: {counts.shape}"
                )
        if counts.shape[2] == 1:
            raise NotImplementedError(
                "Only one isoform provided. Please use a different count model."
            )
        # set model dimensions based on the input shape
        self.n_genes, self.n_spots, self.n_isos = counts.shape

        # switch to float type
        if not counts.dtype.is_floating_point:
            counts = counts.float()

        # convert sparse tensors to dense tensors
        if counts.is_sparse:
            counts = counts.to_dense()

        if design_mtx is None:
            # initialize an empty design matrix of shape (1, n_spots, 0)
            design_mtx = torch.ones(1, self.n_spots, 0)
        elif design_mtx.ndim == 2:
            design_mtx = design_mtx.unsqueeze(0)  # (1, n_spots, n_factors)
        elif design_mtx.ndim == 3:
            # all genes to test should share the same design matrix
            assert (
                design_mtx.shape[0] == 1
            ), "Batched design matrix is currently not supported."
        else:
            raise ValueError(
                f"design_mtx must be a 2D or 3D tensor. Got shape: {design_mtx.shape}"
            )
        self.n_factors = design_mtx.shape[-1]
        assert design_mtx.shape == (
            1,
            self.n_spots,
            self.n_factors,
        ), f"Invalide design matrix shape: {design_mtx.shape}"

        # (n_genes, n_spots, n_isos), int, the observed counts
        self.register_buffer("counts", counts)
        # (1, n_spots, n_factors), the input design matrix of n_factors covariates
        self.register_buffer("X_spot", design_mtx)

        # either the corr_sp or the corr_sp_eigvals and corr_sp_eigvecs must be provided
        assert (corr_sp is not None) or (
            corr_sp_eigvals is not None and corr_sp_eigvecs is not None
        )

        if corr_sp is not None:
            # ignore eigendecomposition if corr_sp is provided
            if corr_sp_eigvals is not None or corr_sp_eigvecs is not None:
                warnings.warn(
                    "Both the correlation matrix and its eigendecomposition are provided. "
                    "The latter will be ignored."
                )

            assert corr_sp.shape == (self.n_spots, self.n_spots)
            # (n_spots, n_spots), the spatial covariance matrix
            self.register_buffer("corr_sp", corr_sp)

            # precompute the eigendecomposition of corr_sp to speed up matrix inverse
            # corr_sp = corr_sp_eigvecs @ diag(corr_sp_eigvals) @ corr_sp_eigvecs.T
            try:
                corr_sp_eigvals, corr_sp_eigvecs = torch.linalg.eigh(self.corr_sp)
            except RuntimeError:
                # fall back to eig if eigh fails
                # related to a pytorch bug on M1 macs, see https://github.com/pytorch/pytorch/issues/83818
                corr_sp_eigvals, corr_sp_eigvecs = torch.linalg.eig(self.corr_sp)
                corr_sp_eigvals = torch.real(corr_sp_eigvals)
                corr_sp_eigvecs = torch.real(corr_sp_eigvecs)

            self.register_buffer("corr_sp_eigvals", corr_sp_eigvals)
            self.register_buffer("corr_sp_eigvecs", corr_sp_eigvecs)

        else:
            assert corr_sp_eigvals.shape == (self.n_spots,)
            assert corr_sp_eigvecs.shape == (self.n_spots, self.n_spots)

            corr_sp = corr_sp_eigvecs.matmul(torch.diag(corr_sp_eigvals)).matmul(
                corr_sp_eigvecs.t()
            )
            # (n_spots, n_spots), the spatial covariance matrix
            self.register_buffer("corr_sp", corr_sp)
            # (n_spots,), the eigenvalues of the spatial covariance matrix
            self.register_buffer("corr_sp_eigvals", corr_sp_eigvals)
            # (n_spots, n_spots), the eigenvectors of the spatial covariance matrix
            self.register_buffer("corr_sp_eigvecs", corr_sp_eigvecs)

        self.register_buffer("convergence", torch.zeros(self.n_genes, dtype=bool))

        # set up learnable parameters according to the model architecture
        self._configure_learnable_variables()
        self._initialize_params()

        # send to device
        self.to(self.device)

    def _configure_learnable_variables(self):
        """Set up learnable parameters according to the model architecture."""
        # the random effect term
        nu = torch.zeros(self.n_genes, self.n_spots, self.n_isos - 1)

        # the fixed effect terms
        beta = torch.zeros(
            self.n_genes, self.n_factors, self.n_isos - 1
        )  # covariate coefficients
        bias_eta = torch.zeros(self.n_genes, self.n_isos - 1)  # intercepts

        # optimize using gradient descent / newton's method
        self.register_parameter("nu", nn.Parameter(nu))
        self.register_parameter("beta", nn.Parameter(beta))
        self.register_parameter("bias_eta", nn.Parameter(bias_eta))

        # the variance components
        n_var_components = 1 if self.share_variance else self.n_isos - 1
        if self.var_parameterization_sigma_theta:
            # cov = sigma^2 (theta * V_sp + (1-theta) * I)
            sigma = torch.ones(self.n_genes, n_var_components)
            theta_logit = torch.zeros(self.n_genes, n_var_components)
            self.register_parameter("sigma", nn.Parameter(sigma))
            self.register_parameter("theta_logit", nn.Parameter(theta_logit))

            if self.var_fix_sigma:
                self.sigma.requires_grad = False
        else:
            # cov = sigma_sp^2 * V_sp + sigma_nsp^2 * I
            sigma_sp = torch.ones(self.n_genes, n_var_components)
            sigma_nsp = torch.ones(self.n_genes, n_var_components)
            self.register_parameter("sigma_sp", nn.Parameter(sigma_sp))
            self.register_parameter("sigma_nsp", nn.Parameter(sigma_nsp))

            if self.var_fix_sigma:
                self.sigma_nsp.requires_grad = False

    def _initialize_params(self):
        """Initialize model parameters."""
        # initialize the random effect term
        if self.init_ratio == "observed":
            # initialize isoform ratios from observed probabilities
            # (n_genes, n_spots, n_isos)
            counts_props = self.counts / (self.counts.sum(-1, keepdim=True) + 1e-5)
            with torch.no_grad():
                self.nu.copy_(
                    (
                        (counts_props[..., :-1] + 1e-5)
                        / (counts_props[..., -1:] + 1e-5)
                    ).log()
                )
        elif self.init_ratio == "uniform":
            # initialize isoforms uniformly across space
            with torch.no_grad():
                self.nu.copy_(torch.zeros(self.n_genes, self.n_spots, self.n_isos - 1))

        # initial estimate of the variance components
        sigma_init = (
            (self.counts.var(1).mean(1) / self.counts.sum(2).mean(1))
            .clamp(max=0.9)
            .pow(0.5)
        ).unsqueeze(-1)

        # initialize proportion of spatial variance to be ~0.05
        if self.var_parameterization_sigma_theta:
            with torch.no_grad():
                self.sigma.copy_(torch.ones_like(self.sigma) * sigma_init)
                self.theta_logit.copy_(torch.ones_like(self.theta_logit) * -3.0)
        else:
            with torch.no_grad():
                self.sigma_nsp.copy_(torch.ones_like(self.sigma_nsp) * sigma_init)
                self.sigma_sp.copy_(torch.ones_like(self.sigma_sp) * 0.2 * sigma_init)

    """Below are a bunch of helper functions to update intermediate variables after each optimization step.
    """

    def var_total(self) -> torch.Tensor:
        """Output the total variance.

        Returns
        -------
        var_total : torch.Tensor
            The total variance ``sigma`` of shape (n_genes, n_var_components).
        """
        if self.var_parameterization_sigma_theta:
            var_total = self.sigma**2
        else:
            var_total = self.sigma_sp**2 + self.sigma_nsp**2

        if var_total.min() < 1e-2:
            warnings.warn("Total variance is close to zero.")
        var_total = torch.clip(var_total, min=1e-2)  # (n_genes, n_var_components)
        return var_total

    def var_sp_prop(self) -> torch.Tensor:
        """Output the proporptions of the spatial variance.

        Returns
        -------
        var_sp_prop : torch.Tensor
            The proportion ``theta`` of spatial variance of shape (n_genes, n_var_components).
        """
        # return: (n_genes, n_var_components)
        if self.var_parameterization_sigma_theta:
            return torch.sigmoid(self.theta_logit)
        else:
            return self.sigma_sp**2 / self.var_total()

    def _corr_eigvals(self):
        """Output the eigenvalues of the correlation matrix."""
        var_sp_prop = self.var_sp_prop().unsqueeze(-1)  # (n_genes, n_var_components, 1)
        # return: (n_genes, n_var_components, n_spots)
        return var_sp_prop * self.corr_sp_eigvals.unsqueeze(0) + (1 - var_sp_prop)
        # return torch.stack(
        #     [(p * self.corr_sp_eigvals + (1 - p)) for p in self.var_sp_prop()], dim=0
        # )

    def _cov_eigvals(self):
        """Output the eigenvalues of the covariance matrix."""
        # (n_genes, n_var_components, n_spots) * (n_genes, n_var_components, 1)
        # return: (n_genes, n_var_components, n_spots)
        return self._corr_eigvals() * self.var_total().unsqueeze(-1)

    def _cov(self):
        """Reconstruct the covariance of the random effect."""
        # return (
        #     self.corr_sp_eigvecs.unsqueeze(0)
        #     @ torch.diag_embed(self._cov_eigvals())
        #     @ self.corr_sp_eigvecs.T.unsqueeze(0)
        # )  # (n_isos - 1) or 1 x n_spots x n_spots
        # return: (n_genes, n_var_components, n_spots, n_spots)
        return self.corr_sp_eigvecs[None, None, ...].matmul(
            torch.diag_embed(self._cov_eigvals()).matmul(
                self.corr_sp_eigvecs.transpose(0, 1)[None, None, ...]
            )
        )

    def _inv_cov(self):
        """Reconstruct the inverse covariance of the random effect."""
        # return (
        #     self.corr_sp_eigvecs.unsqueeze(0)
        #     @ torch.diag_embed(1 / self._cov_eigvals())
        #     @ self.corr_sp_eigvecs.T.unsqueeze(0)
        # )  # (n_isos - 1) or 1 x n_spots x n_spots
        # return: (n_genes, n_var_components, n_spots, n_spots)
        return self.corr_sp_eigvecs[None, None, ...].matmul(
            torch.diag_embed(1 / self._cov_eigvals()).matmul(
                self.corr_sp_eigvecs.transpose(0, 1)[None, None, ...]
            )
        )

    def _eta(self):
        """Output the eta based on the linear model."""
        # eta = X @ beta + bias_eta + nu
        # eta.shape = (n_genes, n_spots, n_isos - 1)
        return self.nu + self.X_spot.matmul(self.beta) + self.bias_eta.unsqueeze(1)

    # Functions to calculate log likelihoods.

    def _calc_log_prob_prior_sigma(self):
        """Calculate log prob of the prior on sigma."""
        if self.var_prior_model == "inv_gamma":  # prior on sigma^2
            return self.var_prior_model_dist.log_prob(self.var_total()).sum(-1)
        elif self.var_prior_model == "gamma":  # prior on sigma
            return self.var_prior_model_dist.log_prob(self.var_total().pow(0.5)).sum(-1)
        else:
            return torch.zeros(self.n_genes, device=self.device)

    def _calc_log_prob_joint(self):
        """Calculate log joint probability given data."""
        # add prior prob of sigma_total
        log_prob = self._calc_log_prob_prior_sigma()  # (n_genes,)

        # add mvn prob of nu ~ MVN(0, S)
        data = self.nu.transpose(1, 2)  # (n_genes, n_isos - 1, n_spots)
        # use the same variance components for all genes if cov_eigvals.shape[1] == 1
        cov_eigvals = self._cov_eigvals().expand(
            self.n_genes, data.shape[1], self.n_spots
        )
        cov_eigvecs = self.corr_sp_eigvecs.expand(
            1, data.shape[1], self.n_spots, self.n_spots
        )
        log_prob += log_prob_fastmvn_batched(
            locs=torch.zeros_like(data),
            cov_eigvals=cov_eigvals,
            cov_eigvecs=cov_eigvecs,
            data=data,
        )

        # add Multinomial likelihood of the counts
        log_prob += log_prob_fastmult_batched(
            self._alpha().transpose(1, 2), self.counts.transpose(1, 2)
        )

        return log_prob  # (n_genes,)

    def _calc_log_prob_marginal(self):
        """Calculate the log marginal probability (integrating out random effect nu)."""
        # by Laplace approximation, the log marginal probability is given by the following:
        # log_marginal ~= log_joint - 1/2 logdet_neg_hessian
        log_prob = self._calc_log_prob_joint()  # (n_genes,)
        full_hessian = (
            self._get_log_lik_hessian_nu()
        )  # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))

        # save the cholesky for fast matrix inverse in Newton updates
        # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        self._chol_hessian_nu = torch.linalg.cholesky(-full_hessian)
        logdet_neg_hessian = 2 * torch.diagonal(
            self._chol_hessian_nu, dim1=-2, dim2=-1
        ).log().sum(
            -1
        )  # (n_genes,)

        return log_prob - 0.5 * logdet_neg_hessian  # (n_genes,)

    def forward(self) -> torch.Tensor:
        """Calculate the log-likelihood or log-marginal-likelihood of the model.

        Returns
        -------
        log_prob : torch.Tensor
            Shape (n_genes,), the log probability for each gene.
        """
        if self.fitting_method in ["marginal_gd", "marginal_newton"]:
            # calculate log marginal prob
            return self._calc_log_prob_marginal()  # (n_genes,)
        else:
            # calculate log joint prob given data
            return self._calc_log_prob_joint()  # (n_genes,)

    # Functions to calculate Hessian.

    def _get_log_lik_hessian_nu(self):
        """Get the Hessian matrix of the log joint probability wrt the random effect nu."""
        n_genes, n_isos = self.n_genes, self.n_isos
        # full_hessian = multinom_hessian + mvn_hessian # n_spots * (n_isos - 1) x n_spots * (n_isos - 1)
        # mvn_hessian: (n_isos - 1) x n_spots x n_spots
        # mvn_hessian[q] = - cov[q]^-1 # n_spots * n_spots
        mvn_hessian = -self._inv_cov()  # (n_genes, n_var_components, n_spots, n_spots)
        if self.share_variance:  # the same variance across isoforms
            # (n_genes, n_isos - 1, n_spots, n_spots)
            mvn_hessian = mvn_hessian.expand(n_genes, n_isos - 1, -1, -1)
        # reshape the hessian into (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        full_hessian = torch.stack(
            [torch.block_diag(*[_m for _m in m_gene]) for m_gene in mvn_hessian], dim=0
        )

        # multinom_hessian[s] = - sum(counts[s,:]) * {diag(p) - p@p.T}
        # self._get_log_lik_hessian_eta() already reshapes the hessian into
        # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        multinom_hessian = self._get_log_lik_hessian_eta()
        full_hessian += multinom_hessian

        return full_hessian

    def _calc_log_prob_mvn_wrt_sigma(self, sigma_expand):
        """Helper function of log joint probability wrt sigmas for calculate Hessian.

        Parameters
        ----------
        sigma_expand : torch.Tensor
            Shape (n_genes, n_var_components, 2). The two variance parameters,
            either (sigma, theta) or (sigma_sp, sigma_nsp), are stacked along the last dimension.
        """
        if self.share_variance:  # the same variance components across isoforms
            sigma_expand = sigma_expand.expand(self.n_genes, self.n_isos - 1, -1)

        if self.var_parameterization_sigma_theta:
            sigma, theta_logit = sigma_expand[..., 0], sigma_expand[..., 1]
            var_sp_prop = torch.sigmoid(theta_logit).unsqueeze(
                -1
            )  # (n_genes, n_iso - 1, -1)
            sigma_total = sigma  # (n_genes, n_iso - 1)
            # (n_genes, n_iso - 1, n_spots) * (n_genes, n_iso - 1, 1) -> (n_genes, n_iso - 1, n_spots)
            cov_eigvals = (
                var_sp_prop * self.corr_sp_eigvals.unsqueeze(0) + (1 - var_sp_prop)
            ) * sigma_total.unsqueeze(-1)
        else:
            sigma_sp, sigma_nsp = sigma_expand[..., 0], sigma_expand[..., 1]
            sigma_total = torch.sqrt(sigma_sp**2 + sigma_nsp**2)  # (n_genes, n_iso - 1)
            # (n_genes, n_iso - 1, 1) * (1, n_spots) -> (n_genes, n_iso - 1, n_spots)
            cov_eigvals = sigma_sp.unsqueeze(-1).pow(
                2
            ) * self.corr_sp_eigvals.unsqueeze(0) + sigma_nsp.unsqueeze(-1).pow(2)

        # MVN prior likelihood as a function of the input eta
        data = self.nu.transpose(1, 2)  # (n_genes, n_isos - 1, n_spots)
        # use the same variance components for all genes if cov_eigvals.shape[1] == 1
        cov_eigvecs = self.corr_sp_eigvecs.expand(
            1, data.shape[1], self.n_spots, self.n_spots
        )
        log_prob = log_prob_fastmvn_batched(
            locs=torch.zeros_like(data),
            cov_eigvals=cov_eigvals,
            cov_eigvecs=cov_eigvecs,
            data=data,
        )

        # add prior prob of sigma_total
        if self.var_prior_model == "inv_gamma":  # prior on sigma^2
            log_prob += self.var_prior_model_dist.log_prob(sigma_total**2).sum(-1)
        elif self.var_prior_model == "gamma":  # prior on sigma
            log_prob += self.var_prior_model_dist.log_prob(sigma_total.abs()).sum(-1)

        return log_prob.mean()

    def _get_sum_of_grad_log_prob_mvn_wrt_sigma(self, sigma_expand):
        """Get the sum of gradients of the log joint probability wrt the variance components."""
        log_prob = self._calc_log_prob_mvn_wrt_sigma(sigma_expand)
        # sum over the batch dim to get shape of (n_var_components, 2)
        return torch.autograd.grad(log_prob, sigma_expand, create_graph=True)[0].sum(0)

    def _get_log_lik_hessian_sigma_expand(self):
        """Get the Hessian matrix of the log mvn wrt the variance components."""
        if self.var_parameterization_sigma_theta:
            sigma_expand = torch.stack(
                [self.sigma, self.theta_logit], dim=-1
            )  # (n_genes, n_var_components, 2)
        else:
            sigma_expand = torch.stack(
                [self.sigma_sp, self.sigma_nsp], dim=-1
            )  # (n_genes, n_var_components, 2)

        # calculate the hessian using pytorch's functional
        # x = sigma_expand.clone().detach().requires_grad_(True)
        hessian_sigma_expand = jacobian(
            self._get_sum_of_grad_log_prob_mvn_wrt_sigma, sigma_expand, vectorize=True
        ).permute(
            2, 0, 1, 3, 4
        )  # (n_genes, n_var_components, 2, n_genes, n_var_components, 2)

        # hessian_sigma_expand = hessian(
        #     self._calc_log_prob_mvn_wrt_sigma, sigma_expand
        # )  # (n_genes, n_var_components, 2, n_genes, n_var_components, 2)

        n_var_comps = hessian_sigma_expand.shape[-2]
        hessian_sigma_expand = hessian_sigma_expand.reshape(
            self.n_genes, n_var_comps * 2, n_var_comps * 2
        )  # (n_genes, n_var_components * 2, n_var_components * 2)

        return hessian_sigma_expand

    """Optimization functions.
    """

    def _update_joint_sigma_expand_newton(self, return_variables=False):
        """Update the variance components by Newton's method.

        Need to first backpropagate the gradients of the log likelihood w.r.t the variance components.
        For concave functions like f(x) = - log(1/x), Newton's update will give worse results.
        """
        n_genes = self.n_genes

        # calculate the updates for variance parameters
        if self.var_parameterization_sigma_theta:
            sigma_expand = torch.stack(
                [self.sigma, self.theta_logit], dim=-1
            )  # (n_genes, n_var_components, 2)
            sigma_expand_grad = torch.stack(
                [self.sigma.grad, self.theta_logit.grad], dim=-1
            )  # (n_genes, n_var_components, 2)
        else:
            sigma_expand = torch.stack(
                [self.sigma_sp, self.sigma_nsp], dim=-1
            )  # (n_genes, n_var_components, 2)
            sigma_expand_grad = torch.stack(
                [self.sigma_sp.grad, self.sigma_nsp.grad], dim=-1
            )  # (n_genes, n_var_components, 2)

        # TODO: analytical solutions when var_prior_model in ['none', 'inv_gamma']?
        hessian_sigma_expand = (
            -self._get_log_lik_hessian_sigma_expand()
        )  # (n_genes, n_var_components * 2, n_var_components * 2)
        hessian_sigma_expand += 1e-5 * torch.eye(
            hessian_sigma_expand.shape[-1]
        ).unsqueeze(0).to(
            self.device
        )  # for stability
        right = hessian_sigma_expand.matmul(
            sigma_expand.reshape(n_genes, -1, 1)
        ) - sigma_expand_grad.reshape(
            n_genes, -1, 1
        )  # (n_genes, n_var_components * 2, 1)
        sigma_expand_new = torch.linalg.solve(
            hessian_sigma_expand, right
        )  # (n_genes, n_var_components * 2, 1)
        sigma_expand_new = sigma_expand_new.reshape(
            sigma_expand.shape
        )  # (n_genes, n_var_components, 2)

        # update the parameters and clear the gradients
        with torch.no_grad():
            if self.var_parameterization_sigma_theta:
                self.sigma.copy_(sigma_expand_new[..., 0])
                self.theta_logit.copy_(sigma_expand_new[..., 1])
                self.sigma.grad.zero_()
                self.theta_logit.grad.zero_()
            else:
                self.sigma_sp.copy_(sigma_expand_new[..., 0])
                self.sigma_nsp.copy_(sigma_expand_new[..., 1])
                self.sigma_sp.grad.zero_()
                self.sigma_nsp.grad.zero_()

        if return_variables:
            return sigma_expand_new

    def _update_joint_newton(self, return_variables=False):
        """Calculate the Newton update of the fixed and random effects.

        For a given parameter p, the Newton update is given by:
                p_new = p - step * hessian^(-1) * gradient

        Returns: (if return_variables)
                nu_new: tensor(n_genes, n_spots, n_isos - 1)
                beta_new: tensor(n_genes, n_factors, n_isos - 1)
                bias_eta_new: tensor(n_genes, n_isos - 1)
                sigma_expand_new: tensor(n_genes, n_var_components, 2)
        """
        n_genes, n_spots, n_isos, n_factors = (
            self.n_genes,
            self.n_spots,
            self.n_isos,
            self.n_factors,
        )
        step = 1  # step size of each update

        # update the random effect term nu
        # self._get_log_lik_hessian_nu() returns the hessian of the log joint likelihood
        # to minize loss (neg likelihood), need to take the negative
        hessian_nu = (
            -self._get_log_lik_hessian_nu()
        )  # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        hessian_nu += 1e-5 * torch.eye(hessian_nu.shape[-1]).unsqueeze(0).to(
            self.device
        )  # for stability
        gradient_nu = self.nu.grad.transpose(1, 2).reshape(
            n_genes, -1, 1
        )  # (n_genes, n_spots * (n_isos - 1), 1)
        right = (
            hessian_nu.matmul(self.nu.transpose(1, 2).reshape(n_genes, -1, 1))
            - step * gradient_nu
        )  # (n_genes, n_spots * (n_isos - 1), -1)
        nu_new = (
            torch.linalg.solve(hessian_nu, right)
            .reshape(n_genes, n_isos - 1, n_spots)
            .transpose(1, 2)
        )
        # update the parameters and clear the gradients
        with torch.no_grad():
            self.nu.copy_(nu_new)
            self.nu.grad.zero_()

        # update the fixed effect term beta and bias
        # self._get_log_lik_hessian_beta_bias() returns the hessian of the log joint likelihood
        # to minize loss (neg likelihood), need to take the negative
        hessian_beta_expand = (
            -self._get_log_lik_hessian_beta_bias()
        )  # (n_genes, (n_factors + 1) * (n_isos - 1), (n_factors + 1) * (n_isos - 1))
        hessian_beta_expand += 1e-5 * torch.eye(
            hessian_beta_expand.shape[-1]
        ).unsqueeze(0).to(self.device)

        # combine beta and bias_eta
        # # (n_genes, n_factors + 1, n_isos - 1)
        # gradient_beta_expand = torch.cat(
        #     [self.beta.grad, self.bias_eta.grad.unsqueeze(1)], dim=1
        # )
        # (n_genes, n_factors + 1, n_isos - 1)
        beta_expand = torch.cat([self.beta, self.bias_eta.unsqueeze(1)], dim=1)
        right = hessian_beta_expand.matmul(
            beta_expand.transpose(1, 2).reshape(n_genes, -1, 1)
        ) - step * beta_expand.transpose(1, 2).reshape(
            n_genes, -1, 1
        )  # (n_genes, (n_factors + 1) * (n_isos - 1), 1)
        beta_expand_new = (
            torch.linalg.solve(hessian_beta_expand, right)
            .reshape(n_genes, n_isos - 1, n_factors + 1)
            .transpose(1, 2)
        )  # (n_genes, n_factors + 1, n_isos - 1)

        # extract beta and bias_eta
        beta_new = beta_expand_new[:, :-1, :]  # (n_genes, n_factors, n_isos - 1)
        bias_eta_new = beta_expand_new[:, -1, :]  # (n_genes, n_isos - 1)
        # update the parameters and clear the gradients
        with torch.no_grad():
            self.beta.copy_(beta_new)
            self.bias_eta.copy_(bias_eta_new)
            self.beta.grad.zero_()
            self.bias_eta.grad.zero_()

        # update variance components
        sigma_expand_new = self._update_joint_sigma_expand_newton(
            return_variables=return_variables
        )

        if return_variables:
            return nu_new, beta_new, bias_eta_new, sigma_expand_new

    def _update_marginal_nu_newton(self, return_variables=False):
        """Calculate the Newton update of the random effects given the hessian."""
        try:
            # use the stored hessian cholesky computed by self._calc_log_prob_margin()
            chol = (
                self._chol_hessian_nu
            )  # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        except AttributeError:
            # if cholesky is not available, compute it on the fly
            chol = torch.linalg.cholesky(-self._get_log_lik_hessian_nu())

        n_genes, n_spots, n_isos = self.n_genes, self.n_spots, self.n_isos
        step = 1  # step size

        hessian_nu_inv = torch.cholesky_inverse(
            chol
        )  # (n_genes, n_spots * (n_isos - 1), n_spots * (n_isos - 1))
        gradient_nu = self.nu.grad.transpose(1, 2).reshape(
            n_genes, -1, 1
        )  # (n_genes, n_spots * (n_isos - 1), 1)
        nu_new = self.nu.transpose(1, 2).reshape(n_genes, -1, 1) - (
            step * hessian_nu_inv
        ).matmul(
            gradient_nu
        )  # (n_genes, n_spots * (n_isos - 1), 1)
        nu_new = nu_new.reshape(n_genes, n_isos - 1, n_spots).transpose(1, 2)
        # update the parameters and clear the gradients
        with torch.no_grad():
            self.nu.copy_(nu_new)
            self.nu.grad.zero_()

        if return_variables:
            return nu_new

    def _fit(
        self,
        diagnose: bool = False,
        verbose: bool = False,
        quiet: bool = False,
        random_seed=None,
    ):
        """Main fitting function to find the MAP estimates using specified fitting method."""
        # extract configs
        fitting_method = self.fitting_method
        optim = self.fitting_configs["optim"]
        max_epochs = self.fitting_configs["max_epochs"]
        patience = self.fitting_configs["patience"]
        tol = self.fitting_configs["tol"]
        max_epochs = 10000 if max_epochs == -1 else max_epochs
        patience = patience if patience > 0 else 1

        if random_seed is not None:  # set random seed for reproducibility
            torch.manual_seed(random_seed)

        # set iteration limits
        epoch = 0
        batch_size = self.n_genes

        # set iteration limits
        batch_size = self.n_genes
        t_start = timer()
        logger = PatienceLogger(batch_size, patience, min_delta=tol, diagnose=diagnose)

        # start training
        self.train()

        while logger.epoch < max_epochs and not logger.convergence.all():
            self.optimizer.zero_grad()

            # minimize the negative log-likelihood or the negative log-marginal-likelihood
            neg_log_prob = -self()  # (n_genes,)
            neg_log_prob.mean().backward()  # backpropate gradients

            if fitting_method == "joint_newton":
                # update nu, beta, bias, and sigmas using Newton's method
                self._update_joint_newton()

            elif fitting_method == "marginal_newton":
                # update nu every k epochs
                if epoch % self.fitting_configs["update_nu_every_k"] == 0:
                    self._update_marginal_nu_newton()
                # skip gradient descent update
                self.nu.grad.zero_()

            # gradient-based updates
            if optim == "lbfgs":
                # update all parameters using L-BFGS
                self.optimizer.zero_grad()
                neg_log_prob = self.optimizer.step(self._closure)
            else:
                # update the remaining parameters with non-zero gradients using gradient descent
                self.optimizer.step()

            # check convergence
            with torch.no_grad():
                # calculate the negative log-likelihood
                neg_log_prob = -self().detach().cpu()

            logger.log(
                neg_log_prob,
                {
                    k: v.detach().cpu()
                    for k, v in self.named_parameters()
                    if "_fake" not in k
                },
            )
            self.convergence.copy_(logger.convergence)

            if (verbose and not quiet) and logger.epoch % 10 == 0:
                print(
                    f"Epoch {logger.epoch}. Loss (neg_log_prob): {logger.best_loss.mean():.4f}. "
                )

        # make sure constraints are satisfied
        self._final_sanity_check()

        # check model convergence
        num_not_converge = (~logger.convergence).sum()
        if num_not_converge:
            warnings.warn(
                f"{num_not_converge} Genes did not converge after epoch {max_epochs}. "
                "Try larger max_epochs."
            )

        # save runtime
        t_end = timer()
        self.fitting_time = t_end - t_start

        if not quiet:  # print final message
            print(
                f"Time {self.fitting_time:.2f}s. Total epoch {logger.epoch}. Final loss "
                f"(neg_log_prob): {neg_log_prob.mean():.3f}."
            )

        # collect parameters corresponding to the best epoch for each sample in batch
        if max_epochs > 0:
            for k, v in self.named_parameters():
                if "_fake" not in k:
                    v.data.copy_(logger.best_params[k])

        self.logger = logger

        return logger.params_iter

    @torch.no_grad()
    def _final_sanity_check(self):
        """Make sure constraints are satisfied."""
        # ensure positive parameters
        if self.var_parameterization_sigma_theta:
            self.sigma.abs_()
        else:
            self.sigma_sp.abs_()
            self.sigma_nsp.abs_()

    def fit(
        self,
        diagnose: bool = False,
        verbose: bool = False,
        quiet: bool = False,
        random_seed: Optional[int] = None,
    ) -> Optional[dict]:
        """Fit the model using all data

        Parameters
        ----------
        diagnose
            Whether to store parameter changes during training (passed to PatienceLogger).
        verbose
            Whether to print verbose information during fitting.
        quiet
            Whether to suppress output during fitting.
        random_seed
            Random seed for reproducibility.

        Returns
        -------
        params_iter : dict or None
            If `diagnose` is True, returns a dictionary of parameter changes during training. Otherwise returns None.
        """
        self._configure_optimizer(verbose=verbose)
        diag_outputs = self._fit(
            diagnose=diagnose, verbose=verbose, quiet=quiet, random_seed=random_seed
        )
        if diagnose:
            return diag_outputs

    def clone(self) -> "MultinomGLMM":
        """Clone a Multinomial GLMM model with the same set of parameters."""
        new_model = type(self)(
            share_variance=self.share_variance,
            var_parameterization_sigma_theta=self.var_parameterization_sigma_theta,
            var_fix_sigma=self.var_fix_sigma,
            var_prior_model=self.var_prior_model,
            var_prior_model_params=self.var_prior_model_params,
            init_ratio=self.init_ratio,
            fitting_method=self.fitting_method,
            fitting_configs=self.fitting_configs,
        )
        new_model.setup_data(
            counts=self.counts,
            corr_sp=None,
            design_mtx=self.X_spot,
            corr_sp_eigvals=self.corr_sp_eigvals,
            corr_sp_eigvecs=self.corr_sp_eigvecs,
        )
        new_model.load_state_dict(self.state_dict())

        return new_model
